import logging
import random
from copy import deepcopy
from typing import List, Tuple

import numpy as np

from salmon.backend.sampler import Sampler
from salmon.triplets.samplers.utils import Answer, Query

logger = logging.getLogger(__name__)


def _get_query(n, head) -> Tuple[int, int, int]:
    a = head
    while True:
        b, c = np.random.choice(n, size=2)
        if a != b and b != c and c != a:
            break
    return a, b, c


def _score_query(q: Tuple[int, int, int]) -> float:
    h, l, r = q
    score = max(abs(h - l), abs(h - r))
    return float(score)


class _RoundRobin(Sampler):
    """
    Let the head of the triplet query rotate through the available items while choosing the bottom two items randomly.
    """

    def __init__(self, n, d=2, ident="", targets=None):
        """
        Parameters
        ----------
        n : int
            Number of objects
        ident : str
            Identifier of the algorithm
        targets : Optional[List[int]]
            The allowable indexes to ask about.
        """
        self.n = n
        self.d = d
        self.answers = []
        self.counter = 0

        self.order = None
        self.targets = targets or list(range(n))

        if targets is not None:
            if not isinstance(targets, list):
                msg = "Specify a list for targets. Got {} or type {}"
                raise ValueError(msg.format(targets, type(targets)))
            if not all(isinstance(i, int) for i in targets):
                msg = "Not all items in targets are integers. Bad values are {}"
                bad_vals = [v for v in targets if not isinstance(v, int)]
                raise ValueError(msg.format(bad_vals))
            if len(targets) < 3:
                msg = "Specify at least 3 targets items. Got {} targets items"
                raise ValueError(msg.format(len(targets)))
            if max(targets) >= n:
                msg = "At least one targets target is too large. Values too large include {}, larger than {}"
                bad_vals = [v for v in targets if v >= n]
                raise ValueError(msg.format(bad_vals, n))

        super().__init__(ident=ident)

    def get_query(self, **kwargs) -> Tuple[Query, float]:
        if self.order is None:
            self.order = deepcopy(self.targets)

        idx = self.counter % len(self.order)
        logger.warning("idx=%s", idx)
        if idx == 0:
            np.random.shuffle(self.order)

        head = self.order[idx]

        kids = set(self.targets) - {head}
        a, b = np.random.choice(list(kids), size=2, replace=False)
        self.counter += 1
        score = max(abs(head - a), abs(head - b))
        return {"head": int(head), "left": int(a), "right": int(b)}, float(score)

    def process_answers(self, ans: List[Answer]):
        return self, False

    def run(self, *args, **kwargs):
        from rejson import Path
        root = Path.rootPath()

        rj = self.redis_client()
        rj.jsonset(f"stopped-{self.ident}", Path("."), True)
        return None


class RoundRobin(_RoundRobin):
    """
    Let the head of the triplet query rotate through the available items while choosing
    the bottom two items randomly. This class is user specific if the
    ``/query?puid=foo`` endpoint is hit.
    """
    def __init__(self, *args, **kwargs):
        self.rr_args = args
        self.rr_kwargs = kwargs
        self.samplers = {}  # puid to roundrobing

        super().__init__(*args, **kwargs)

    def get_query(self, puid: str = "") -> Tuple[Query, float]:
        if puid not in self.samplers:
            self.samplers[puid] = _RoundRobin(*self.rr_args, **self.rr_kwargs)
        return self.samplers[puid].get_query()

    def process_answers(self, ans: List[Answer]):
        return self, True



from pathlib import Path
import yaml
import random
from typing import Dict

import numpy as np
import numpy.linalg as LA
import pandas as pd
import pytest
from sklearn.model_selection import train_test_split

from salmon.triplets.offline import OfflineEmbedding
from salmon.triplets.samplers import TSTE
import salmon.triplets.offline

ArrayLike = np.ndarray

def test_salmon_import():
    """This test makes sure that no errors are raised on import
    (non-existant directories, etc)"""
    import salmon

    return True


def test_score_predict_basic():
    """Test the interface for score/predict"""
    n = 10
    n_ans = 40
    X = np.random.choice(n, size=(n_ans, 3)).astype("uint8")
    alg = TSTE(n=n, d=2)

    # Test expected interface
    y = np.random.choice(2, size=n_ans).astype("uint8")
    y_hat = alg.predict(X)
    assert y_hat.shape == (n_ans,)
    assert np.unique(y_hat).tolist() == [0, 1]

    acc = alg.score(X, y)
    assert isinstance(acc, float)
    assert 0 <= acc <= 1


def test_score_accurate():
    n = 10
    n_ans = 40
    X = np.random.choice(n, size=(n_ans, 3)).astype("uint8")
    alg = TSTE(n=n, d=2)
    y_hat = alg.predict(X)

    # Make sure perfect accuracy if the output aligns with the embedding
    assert np.allclose(alg.score(X, y_hat), 1)

    # Make sure the score has the expected value (winner has minimum distance)
    embed = alg.opt.embedding() * 1e3
    y_hat2 = []
    for head, left, right in X:
        ldist = LA.norm(embed[head] - embed[left])
        rdist = LA.norm(embed[head] - embed[right])

        left_wins = ldist <= rdist
        y_hat2.append(0 if left_wins else 1)

    assert y_hat.tolist() == y_hat2


def test_offline_embedding():
    n, d = 85, 2
    max_epochs = 20

    X = np.random.choice(n, size=(10_000, 3))
    X_train, X_test = train_test_split(X, random_state=0, test_size=0.2)

    model = OfflineEmbedding(n=n, d=d, max_epochs=max_epochs)
    model.fit(X_train, X_test)
    assert isinstance(model.embedding_, np.ndarray)
    assert model.embedding_.shape == (n, d)

    assert isinstance(model.history_, list)
    assert all(isinstance(h, dict) for h in model.history_)
    epochs = model.history_[-1]["num_grad_comps"] / len(X_train)
    eps = 0.75
    assert max_epochs - eps <= epochs <= max_epochs + eps


def test_offline_embedding_random_state():
    n, d = 85, 2
    max_epochs = 20
    random_state = 20

    X = np.random.choice(n, size=(10_000, 3))
    X_train, X_test = train_test_split(X, random_state=0, test_size=0.2)

    m1 = OfflineEmbedding(
        n=n, d=d, max_epochs=max_epochs, random_state=random_state
    ).initialize(X_train)
    m2 = OfflineEmbedding(
        n=n, d=d, max_epochs=max_epochs, random_state=random_state
    ).initialize(X_train)
    assert np.allclose(m1.embedding_, m2.embedding_)


def test_offline_init():
    n, d = 20, 2

    X = np.random.choice(n, size=(100, 3))
    em = np.random.uniform(size=(n, d))
    est = OfflineEmbedding(n=n, d=d)
    est.initialize(X, embedding=em)

    assert np.allclose(est.embedding_, em)
    est.partial_fit(X)
    assert not np.allclose(est.embedding_, em), "Embedding didn't change"


def test_offline_names_correct():
    DIR = Path(__file__).absolute().parent
    _f = DIR / "data" / "active.yaml"
    config = yaml.load(_f.read_text(), Loader=yaml.SafeLoader)
    n = len(config["targets"])
    d = config["sampling"]["common"]["d"]

    X = np.random.choice(n, size=(100, 3))
    est = OfflineEmbedding(n=n, d=d)
    est.partial_fit(X)

    import salmon.triplets.offline as offline

    em = offline.join(est.embedding_, config["targets"])
    assert isinstance(em, pd.DataFrame)
    assert len(em) == len(config["targets"])
    assert set(em.columns) == {"x", "y", "target"}
    assert (em["target"] == config["targets"]).all()


def _answer(q: Dict[str, int], X: ArrayLike) -> int:
    h = X[q["head"]]
    l = X[q["left"]]
    r = X[q["right"]]
    if LA.norm(h - l) < LA.norm(h - r):
        return q["left"]
    return q["right"]

def test_offline_adaptive(n=10, d=1):
    rng = np.random.RandomState(42)
    X = rng.uniform(size=(n, d))
    val_queries = np.asarray([rng.choice(n, size=3, replace=False) for _ in range(5000)])
    val_ans = np.asarray([0 if LA.norm(X[h] - X[l]) < LA.norm(X[h] - X[r]) else 1 for h, l, r in val_queries])

    sampler = TSTE(n=n, d=d, alpha=1, R=1)
    score0 = sampler.score(val_queries, val_ans)
    for t in range(20):
        queries, scores, _ = sampler.get_queries()
        _good_queries = queries[np.argsort(scores)[-5:]]
        good_queries = [{"head": h, "left": o1, "right": o2} for h, o1, o2 in _good_queries]
        answers = [{"winner": _answer(q, X), **q} for q in good_queries]
        sampler.process_answers(answers)

    scorem1 = sampler.score(val_queries, val_ans)
    assert 0 <= score0 <= scorem1 <= 1
    assert score0 + 0.05 < scorem1, "Improves by at least 5% after 100 answers (often much larger)"


if __name__ == "__main__":
    test_offline_init()
    test_offline_embedding_random_state()
    test_offline_embedding()



"""
utilsCrowdKernel.py

author: Lalit Jain (lalitkumarj@gmail.com)
edited: 1/18/15

This module has methods that assist with the crowd kernel for pairwise comparisons presented here: 
http://research.microsoft.com/en-us/um/people/adum/publications/2011-crowd-kernel.pdf

If you're trying to COMPUTE an embedding, you might simply call:
    X,emp_loss = computeEmbedding(n,d,S)

You may also consider getLoss to check how well an embedding is performing.
"""

from __future__ import print_function

import numpy.random
from numpy import *
from numpy.linalg import *
from numpy.random import *

# eig = numpy.linalg
norm = linalg.norm
floor = math.floor
ceil = math.ceil
realmin = finfo(double).tiny
log = math.log

import time


def main():
    """
    Example of Usage

    Creates some fake data and finds an embedding
    """

    # generate some fake data
    n = 30
    d = 2
    m = int(ceil(40 * n * d * log(n)))  # number of labels

    p = 0.1
    # error rate

    Strain = []
    Stest = []
    Xtrue = randn(n, d)
    for iter in range(0, m):

        # get random triplet
        q, score = getRandomQuery(Xtrue)

        # align it so it agrees with Xtrue: "q[2] is more similar to q[0] than q[1]"
        # q[2]: winner
        # q[1]: loser
        # q[0]: head
        query_ordering_disagrees_with_Xtrue = score < 0
        if query_ordering_disagrees_with_Xtrue:
            q = [q[i] for i in [1, 0, 2]]

        # add some noise
        R = rand()
        if R < p:
            q = [q[i] for i in [1, 0, 2]]

        if iter < 0.9 * m:
            Strain.append(q)
        else:
            Stest.append(q)

    alpha = 1.0
    # compute embedding
    X, emp_loss_train = computeEmbedding(
        n, d, Strain, alpha=alpha, num_random_restarts=2, epsilon=0.00001, verbose=True
    )

    # compute loss on test set
    emp_loss_test, hinge_loss_test, log_loss_test = getLoss(X, Stest, alpha)

    print()
    print("Training loss = %f,   Test loss = %f" % (emp_loss_train, emp_loss_test))


def getRandomQuery(X):
    """
    Outputs a triplet [i,j,k] chosen uniformly at random from all possible triplets 
    and score = abs( ||x_i - x_k||^2 - ||x_j - x_k||^2 )

    Inputs:
        (numpy.ndarray) X : matrix from which n is extracted from and score is derived

    Outputs:
        [(int) i, (int) j, (int) k] q : where k in [n], i in [n]-k, j in [n]-k-j
        (float) score : signed distance to current solution (positive if it agrees, negative otherwise)

    Usage:
        q,score = getRandomQuery(X)
    """
    n, d = X.shape

    i = randint(n)
    j = randint(n)
    while j == i:
        j = randint(n)
    k = randint(n)
    while (k == i) | (k == j):
        k = randint(n)
    q = [i, j, k]

    score = getTripletScore(X, q)

    return q, score


def getTripletScore(X, q):
    """
    Given X,q=[i,j,k] returns score = ||x_j - x_k||^2 - ||x_i - x_k||^2
    If score > 0 then the triplet agrees with the embedding, otherwise it does not 

    Usage:
        score = getTripletScore(X,[3,4,5])
    """
    i, j, k = q

    return dot(X[j], X[j]) - 2 * dot(X[j], X[k]) + 2 * dot(X[i], X[k]) - dot(X[i], X[i])


def getLoss(X, S, alpha):
    """
    Returns loss on X with respect to list of triplets S: 1/len(S) \sum_{q in S} loss(X,q).
    Intuitively, q=[i,j,k] "agrees" with X if ||x_j - x_k||^2 > ||x_i - x_k||^2.

    For q=[i,j,k], let s(X,q) = ||x_j - x_k||^2 - ||x_i - x_k||^2
    If loss is hinge_loss then loss(X,q) = max(0,1-s(X,q))
    If loss is emp_loss then loss(X,q) = 1 if s(X,q)<0, and 0 otherwise

    Usage:
        emp_loss, hinge_loss = getLoss(X,S)
    """
    n = X.shape[0]
    d = X.shape[1]

    emp_loss = 0  # 0/1 loss
    hinge_loss = 0  # hinge loss
    log_loss = 0  # log_loss in crowd kernel model
    for q in S:
        loss_ijk = getTripletScore(X, q)
        hinge_loss = hinge_loss + max(0, 1.0 - loss_ijk)
        log_loss = log_loss - log(
            getSTETripletProbability(X[q[0]], X[q[1]], X[q[2]], alpha=1)
        )

        if loss_ijk < 0:
            emp_loss = emp_loss + 1.0

    emp_loss = emp_loss / len(S)
    hinge_loss = hinge_loss / len(S)
    log_loss = log_loss / len(S)
    return emp_loss, hinge_loss, log_loss


def getSTETripletProbability(i, j, k, alpha=1):
    """
    Return the probability of triplet [i,l,j] where a is closer to b than c.

    Namely:
    pabc = (1 + || c - a||^2/alpha)**(-(alpha+1)/2)/(2*alpha + || b - a ||^2+|| c - a ||^2)

    Inputs:
        (numpy.ndarray) a : numpy array
        (numpy.ndarray) b : numpy array
    (numpy.ndarray) c : numpy array
        (float) alpha : regularization parameter

    Notes
    -----
    head : k
    winner : i
    loser : j
    """
    ki = norm(k - i)
    kj = norm(k - j)
    c = -(alpha + 1.0) / 2
    return (1 + ki * ki / alpha) ** c / (
        (1 + ki * ki / alpha) ** c + (1 + kj * kj / alpha) ** c
    )


def getEntropy(tau):
    """
    Return the entropy of a distribution tau.

    Namely:
    H(tau) = sum -tau[i]*log(tau[i])

    Inputs:
        (numpy.ndarray) tau : numpy array representing a probability distribution
    """
    e = 0
    for i in range(len(tau)):
        if tau[i] > 0:
            e += -1 * tau[i] * log(tau[i])
    return e


def getSTETauDistribution(X, S, alpha=1):
    """
    Return the tau distributions for each point [n].

    Returns a len(X)xlen(X) two dimensional arrray. The ith row of a is the tau distribution corresponding to the posterior distribution of a.

    Inputs:
    (numpy.ndarray) X: current embedding of points
    (list) S: list of triplets
    (float) alpha: regularization parameter
    Usage:
        tau = getSTEDistribution(X,S) 
    """
    n, d = X.shape
    tau = zeros((n, n))
    # Loop over each query
    for q in S:
        a = q[2]
        # Multiply by the amount the query contributes to each tau
        for i in range(n):
            tau[a, i] = tau[a, i] + log(
                getSTETripletProbability(X[q[0]], X[q[1]], X[i], alpha=alpha)
            )

    # Normalize -- make each row a PDF
    for a in range(n):
        tau[a] = exp(tau[a])
        s = sum(tau[a])
        tau[a] = tau[a] / s

    return tau


def getGradient(X, S, alpha):
    """
    Returns gradient of the log loss of the crowd kernel probability distribution.
    Requires a parameter alpha for regularization.
    
    Usage:
        G,avg_grad_row_norm_sq,max_grad_row_norm_sq,avg_row_norm_sq = getGradient(X,S)
    """
    n, d = X.shape
    m = len(S)

    G = zeros((n, d))
    c = -(alpha + 1.0) / 2
    # By defintion d = 2c/alpha
    cc = -(alpha + 1.0) / alpha
    log_loss = 0
    for q in S:
        i, j, k = q
        normki = norm(X[k] - X[i])
        normki = normki * normki
        normkj = norm(X[k] - X[j])
        normkj = normkj * normkj
        num = max((1 + normki / alpha) ** c, realmin)
        den = max(
            (1 + normki / alpha) ** c + (1 + normkj / alpha) ** c, realmin
        )  # max(2*alpha + D[k,i] + D[k,j], realmin)
        P = num / den
        G[k] += (
            (1 - P)
            * cc
            * (
                (X[k] - X[i]) / (1 + normki / alpha)
                - (X[k] - X[j]) / (1 + normkj / alpha)
            )
        )
        G[i] += -(1 - P) * cc * (X[k] - X[i]) / (1 + normki / alpha)
        G[j] += (1 - P) * cc * (X[k] - X[j]) / (1 + normkj / alpha)
        log_loss += log(den) - log(num)

    log_loss = log_loss / len(S)
    # Remember, the loss function is the sum of log(1/p^k_ij), this leads to an extra minus sign
    G = -1.0 / len(S) * G
    # compute statistics about gradient used for stopping conditions
    mu = mean(X, 0)
    avg_row_norm_sq = 0.0
    avg_grad_row_norm_sq = 0.0
    max_grad_row_norm_sq = 0.0
    norm_grad_sq_0 = 0.0

    for i in range(n):
        row_norm_sq = 0
        grad_row_norm_sq = 0
        for j in range(d):
            row_norm_sq += (X[i, j] - mu[j]) * (X[i, j] - mu[j])
            grad_row_norm_sq += G[i, j] * G[i, j]
        avg_row_norm_sq += row_norm_sq / n
        avg_grad_row_norm_sq += grad_row_norm_sq / n
        max_grad_row_norm_sq = max(max_grad_row_norm_sq, grad_row_norm_sq)

    return G, log_loss, avg_grad_row_norm_sq, max_grad_row_norm_sq, avg_row_norm_sq


def computeEmbedding(
    n,
    d,
    S,
    alpha=1,
    num_random_restarts=0,
    max_num_passes=0,
    max_norm=0,
    epsilon=0.001,
    verbose=False,
):
    """
    Computes an embedding of n objects in d dimensions usin the triplets of S.
    S is a list of triplets such that for each q in S, q = [i,j,k] means that
    object k should be closer to i than j.

    Inputs:
        (int) n : number of objects in embedding
        (int) d : desired dimension
        (list [(int) i, (int) j,(int) k]) S : list of triplets, i,j,k must be in [n].
        (float) mu : regularization value 
        (int) num_random_restarts : number of random restarts (nonconvex optimization, may converge to local minima)
        (int) max_num_passes : maximum number of passes over data SGD makes before proceeding to GD (default equals 16)
        (float) max_norm : the maximum allowed norm of any one object (default equals 10*d)
        (float) epsilon : parameter that controls stopping condition, smaller means more accurate (default = 0.01)
        (boolean) verbose : outputs some progress (default equals False)

    Outputs:
        (numpy.ndarray) X : output embedding
        (float) gamma : Equal to a/b where a is max row norm of the gradient matrix and b is the avg row norm of the centered embedding matrix X. This is a means to determine how close the current solution is to the "best" solution.  
    """

    if max_num_passes == 0:
        max_num_passes = 32

    X_old = None
    emp_loss_old = float("inf")
    num_restarts = -1
    while num_restarts < num_random_restarts:
        num_restarts += 1

        # print "Epoch SGD"
        # ts = time.time()
        # X,acc = computeEmbeddingWithEpochSGD(n, d, S, alpha, max_num_passes=max_num_passes, max_norm=max_norm, epsilon=0., verbose=verbose)
        # te_sgd = time.time()-ts
        X = randn(n, d) * 0.0001

        # print "Gradient Descent"
        ts = time.time()
        (
            X_new,
            emp_loss_new,
            hinge_loss_new,
            log_loss_new,
            acc_new,
        ) = computeEmbeddingWithGD(
            X,
            S,
            alpha,
            max_iters=50,
            max_norm=max_norm,
            epsilon=epsilon,
            verbose=verbose,
        )
        emp_loss_new, hinge_loss_new, log_loss_new = getLoss(X_new, S, alpha)
        te_gd = time.time() - ts

        if emp_loss_new < emp_loss_old:
            X_old = X_new
            emp_loss_old = emp_loss_new

        if verbose:
            print(
                "restart %d:   emp_loss = %f,   hinge_loss = %f,   duration=%f+%f"
                % (num_restarts, emp_loss_new, hinge_loss_new, te_sgd, te_gd)
            )

    return X_old, emp_loss_old


def computeEmbeddingWithEpochSGD(
    n, d, S, alpha=1, max_num_passes=0, max_norm=0, epsilon=0.01, a0=0.1, verbose=False
):
    """
    Performs epochSGD where step size is constant across each epoch, epochs are 
    doubling in size, and step sizes are getting cut in half after each epoch.
    This has the effect of having a step size decreasing like 1/T. a0 defines 
    the initial step size on the first epoch. 

    S is a list of triplets such that for each q in S, q = [i,j,k] means that
    object k should be closer to i than j.

    Inputs:
        (int) n : number of objects in embedding
        (int) d : desired dimension
        (list [(int) i, (int) j,(int) k]) S : list of triplets, i,j,k must be in [n].
        (float) mu : regularization value 
        (int) max_num_passes : maximum number of passes over data (default equals 16)
        (float) max_norm : the maximum allowed norm of any one object (default equals 10*d)
        (float) epsilon : parameter that controls stopping condition (default = 0.01)
        (float) a0 : inititial step size (default equals 0.1)
        (boolean) verbose : output iteration progress or not (default equals False)

    Outputs:
        (numpy.ndarray) X : output embedding
        (float) gamma : Equal to a/b where a is max row norm of the gradient matrix and b is the avg row norm of the centered embedding matrix X. This is a means to determine how close the current solution is to the "best" solution.  


    Usage:
        X,gamma = computeEmbeddingWithEpochSGD(n,d,S)
    """
    m = len(S)

    # norm of each object is equal to 1 in expectation
    X = randn(n, d) * 0.0001

    if max_num_passes == 0:
        max_iters = 16 * m
    else:
        max_iters = max_num_passes * m

    if max_norm == 0:
        max_norm = 10.0 * d

    epoch_length = m
    a = a0
    t = 0
    t_e = 0

    # check losses
    if verbose:
        emp_loss, hinge_loss, log_loss = getLoss(X, S, alpha)
        print(
            "iter=%d,   emp_loss=%f,   hinge_loss=%f, log_loss=%f,  a=%f"
            % (0, emp_loss, hinge_loss, log_loss, a)
        )
    rel_max_grad = None
    while t < max_iters:
        t += 1
        t_e += 1

        # check epoch conditions, udpate step size
        if t_e % epoch_length == 0:
            a = a * 0.5
            epoch_length = 2 * epoch_length
            t_e = 0

            if epsilon > 0 or verbose:
                # get losses
                emp_loss, hinge_loss, log_loss = getLoss(X, S, alpha)

                # get gradient and check stopping-time statistics
                (
                    G,
                    log_loss,
                    avg_grad_row_norm_sq,
                    max_grad_row_norm_sq,
                    avg_row_norm_sq,
                ) = getGradient(X, S, alpha)
                rel_max_grad = sqrt(max_grad_row_norm_sq / avg_row_norm_sq)
                rel_avg_grad = sqrt(avg_grad_row_norm_sq / avg_row_norm_sq)

                if verbose:
                    print(
                        "iter=%d,   emp_loss=%f,   hinge_loss=%f,  log_loss=%f,   rel_avg_grad=%f,   rel_max_grad=%f,   a=%f"
                        % (
                            t,
                            emp_loss,
                            hinge_loss,
                            log_loss,
                            rel_avg_grad,
                            rel_max_grad,
                            a,
                        )
                    )

                if rel_max_grad < epsilon:
                    break

        # get random triplet uniformly at random
        q = S[randint(m)]
        c = -(alpha + 1.0) / 2
        # By defintion d = 2c/alpha
        cc = -(alpha + 1.0) / alpha
        # take gradient step
        i, j, k = q
        normki = norm(X[k] - X[i]) ** 2
        normkj = norm(X[k] - X[j]) ** 2
        num = (1 + normki / alpha) ** c
        den = (1 + normki / alpha) ** c + (1 + normkj / alpha) ** c
        P = num / den

        grad_k = (
            -(1 - P)
            * cc
            * (
                (X[k] - X[i]) / (1 + normki / alpha)
                - (X[k] - X[j]) / (1 + normkj / alpha)
            )
        )
        grad_i = (1 - P) * cc * (X[k] - X[i]) / (1 + normki / alpha)
        grad_j = -(1 - P) * cc * (X[k] - X[j]) / (1 + normkj / alpha)

        X[i] = X[i] - a * grad_i / len(S)
        X[j] = X[j] - a * grad_j / len(S)
        X[k] = X[k] - a * grad_k / len(S)

        # project back onto ball such that norm(X[i])<=max_norm
        for i in q:
            norm_i = norm(X[i])
            if norm_i > max_norm:
                X[i] = X[i] * (max_norm / norm_i)

    return X, rel_max_grad


def computeEmbeddingWithGD(
    X,
    S,
    alpha=1,
    max_iters=0,
    max_norm=0,
    epsilon=0.001,
    c1=0.0001,
    rho=0.7,
    verbose=False,
):
    """
    Performs gradient descent with step size as implemented in stochastic triplet embedding code, namely ckl_x.m 
    See: http://homepage.tudelft.nl/19j49/ste/Stochastic_Triplet_Embedding_files/STE_Release.zip


    S is a list of triplets such that for each q in S, q = [i,j,k] means that
    object k should be closer to i than j.

    Inputs:
        (numpy.ndarray) X : input embedding
        (list [(int) i, (int) j,(int) k]) S : list of triplets, i,j,k must be in [n]. 
        (float) mu : regularization parameter
        (int) max_iters : maximum number of iterations of SGD (default equals 40*len(S))
        (float) max_norm : the maximum allowed norm of any one object (default equals 10*d)
        (float) epsilon : parameter that controls stopping condition, exits if gamma<epsilon (default = 0.01)
        (float) c1 : Amarijo stopping condition parameter (default equals 0.0001)
        (float) rho : Backtracking line search parameter (default equals 0.5)
        (boolean) verbose : output iteration progress or not (default equals False)

    Outputs:
        (numpy.ndarray) X : output embedding
        (float) emp_loss : output 0/1 error
        (float) hinge_loss : output hinge loss
        (float) gamma : Equal to a/b where a is max row norm of the gradient matrix and b is the avg row norm of the centered embedding matrix X. This is a means to determine how close the current solution is to the "best" solution.  


    Usage:
        X,gamma = computeEmbeddingWithGD(X,S)
    """
    m = len(S)

    n, d = X.shape

    if max_iters == 0:
        max_iters = 16 * m

    if max_norm == 0:
        max_norm = 10 * d

    # check losses
    if verbose:
        emp_loss, hinge_loss, log_loss = getLoss(X, S, alpha)
        print(
            "iter=%d,   emp_loss=%f,   hinge_loss=%f, log_loss=%f,   a=%f"
            % (0, emp_loss, hinge_loss, log_loss, float("nan"))
        )
    a = 0.5 * n
    t = 0
    emp_loss_0 = float("inf")
    hinge_loss_0 = float("inf")
    log_loss_0 = float("inf")
    rel_max_grad = float("inf")

    while t < max_iters:
        t += 1
        # get gradient and stopping-time statistics
        ts = time.time()
        (
            G,
            log_loss,
            avg_grad_row_norm_sq,
            max_grad_row_norm_sq,
            avg_row_norm_sq,
        ) = getGradient(X, S, alpha)
        rel_max_grad = sqrt(max_grad_row_norm_sq / avg_row_norm_sq)
        rel_avg_grad = sqrt(avg_grad_row_norm_sq / avg_row_norm_sq)
        if rel_max_grad < epsilon:
            if verbose:
                print("Exited on rel_max_grad %s" % (str(rel_max_grad)))
            break

        # perform backtracking line search
        a = 2 * a
        ts = time.time()
        emp_loss_0, hinge_loss_0, log_loss_0 = getLoss(X, S, alpha)
        norm_grad_sq_0 = avg_grad_row_norm_sq * n
        emp_loss_k, hinge_loss_k, log_loss_k = getLoss(X - a * G, S, alpha)

        inner_t = 0
        while log_loss_k > log_loss_0 - c1 * a * norm_grad_sq_0:
            a = a * rho
            emp_loss_k, hinge_loss_k, log_loss_k = getLoss(X - a * G, S, alpha)
            inner_t += 1
        X = X - a * G

        for i in range(n):
            norm_i = norm(X[i])
            if norm_i > max_norm:
                X[i] = X[i] * (max_norm / norm_i)

        # check losses
        if verbose:
            print(
                "ste  iter=%d,   emp_loss=%f,   hinge_loss=%f,   log_loss=%f,   rel_avg_grad=%f,   rel_max_grad=%f,   a=%f,   i_t=%d"
                % (
                    t,
                    emp_loss_k,
                    hinge_loss_k,
                    log_loss_k,
                    rel_avg_grad,
                    rel_max_grad,
                    a,
                    inner_t,
                )
            )

    return X, emp_loss_0, hinge_loss_0, log_loss_0, rel_max_grad


if __name__ == "__main__":
    main()



import hashlib
import os
from io import BytesIO
from pathlib import Path
from textwrap import dedent
from typing import Any, Union
from zipfile import ZipFile

import pandas as pd
from fastapi import HTTPException

from salmon.utils import get_logger

logger = get_logger(__name__)


class ServerException(HTTPException):
    def __init__(self, msg, status_code=500):
        logger.error(msg)
        raise HTTPException(status_code=status_code, detail=msg)


def _extract_zipfile(raw_zipfile, directory="targets"):
    p = Path(__file__).absolute().parent  # directory to this file
    imgs = p / "static" / directory

    if imgs.exists():
        for _f in imgs.glob("**/*"):
            _f.unlink()
        imgs.rmdir()
    if not imgs.exists():
        imgs.mkdir()
    with BytesIO(raw_zipfile) as f:
        with ZipFile(f) as myzip:
            infos = [f for f in myzip.infolist() if f.filename[0] != "."]
            for info in infos:
                if info.filename[-1] == "/" or "/." in info.filename:
                    continue
                info.filename = os.path.basename(info.filename)
                myzip.extract(info, path=str(imgs))
    fnames = list(imgs.glob("**/*"))

    def _numeric_fname(f: Path) -> Union[float, str]:
        str_or_digit = str(f.name).split(".")[0]
        return float(str_or_digit) if str_or_digit.isdigit() else f.name

    return list(sorted(fnames, key=_numeric_fname))


def _format_target(file: Path):
    static = Path(__file__).absolute().parent
    p = file.relative_to(static)
    logger.info(str(p))
    url = "/" + str(p)
    if any(ext in url.lower() for ext in ["png", "gif", "jpg", "bmp", "jpeg", "svg"]):
        return f"<img src='{url}' />"
    elif any(ext in url for ext in ["mov", "mp4"]):
        return dedent(
            f"""
            <video autoplay controls>
            <source src='{url}' type='video/mp4' />
            Your browser does not support the video tag.
            </video>
            """
        )
    else:
        raise ValueError(
            f"Unsupported extension for file={file}. "
            "Supported extensions are ['png', 'gif', 'jpg', 'bmp', "
            "'jpeg', 'svg', 'mov' or 'mp4']"
        )


def _format_targets(file: Path):
    df = pd.read_csv(file, header=None)
    if len(df.columns) > 1:
        raise ValueError("Unsupported CSV file. One target should be on each line.")
    return df[df.columns[0]].tolist()


def image_url(target: str) -> str:
    i = target.find("src=")
    t = target[i + 5 :]
    return t[: t.find("'")]


def sha256(x: Any) -> str:
    if not isinstance(x, (str, bytes)):
        x = str(x)
    if isinstance(x, str):
        x = x.encode(encoding="ascii")
    m = hashlib.sha256()
    m.update(x)
    return m.hexdigest()



import json
from logging import getLogger
from pathlib import Path
from time import sleep
from typing import Tuple
from warnings import warn

import httpx as requests
import numpy as np
import pytest
import yaml
from sklearn.utils import check_random_state

logger = getLogger(__name__)
TIMEOUT = 200


class Server:
    def __init__(self, url, async_client=None):
        self.url = url
        self.async_client = async_client
        self._authorized = False

    @property
    def creds(self) -> Tuple[str, str]:
        return ("foobar", "pass123")

    def get(self, endpoint, status_code=200, **kwargs):
        if "auth" not in kwargs and self._authorized:
            kwargs.update(auth=self.creds)
        if "reset" in endpoint and "timeout" not in kwargs:
            kwargs.update(timeout=TIMEOUT)
        if "timeout" not in kwargs:
            kwargs.update(timeout=TIMEOUT)

        logger.info(f"Getting {endpoint}")
        r = requests.get(self.url + endpoint, **kwargs)
        logger.info("done")
        if status_code:
            assert r.status_code == status_code, (r.status_code, status_code, r.text)
        return r

    def reset(self):
        r = self.delete("/reset?force=1", auth=self.creds, timeout=TIMEOUT)
        assert r.json() == {"success": True}

    def post(self, endpoint, data=None, status_code=200, error=False, **kwargs):
        if "auth" not in kwargs and self._authorized:
            kwargs.update(auth=self.creds)
        if "init_exp" in endpoint and "timeout" not in kwargs:
            kwargs.update(timeout=TIMEOUT)
        if "timeout" not in kwargs:
            kwargs.update(timeout=TIMEOUT)
        if isinstance(data, dict) and "exp" not in data and "rdb" not in data:
            data = json.dumps(data)
        logger.info(f"Getting {endpoint}")
        r = requests.post(self.url + endpoint, data=data, **kwargs)
        logger.info("done")
        if not error:
            assert r.status_code == status_code, (r.status_code, status_code, r.text)
        return r

    def delete(self, endpoint, status_code=200, **kwargs):
        if "auth" not in kwargs and self._authorized:
            kwargs.update(auth=self.creds)
        logger.info(f"Getting {endpoint}...")
        r = requests.delete(self.url + endpoint, **kwargs)
        logger.info("done")
        assert r.status_code == status_code, (r.status_code, status_code, r.text)
        return r

    def authorize(self):
        username, password = self.creds
        r = self.post(f"/create_user/{username}/{password}", error=True)
        assert r.status_code in {200, 403}
        self._authorized = True
        return r


def _clear_logs(log=None):
    if log:
        with log.open(mode="w") as f:
            print("", file=f)
    else:
        this_dir = Path(__file__).absolute().parent
        root_dir = this_dir.parent
        log_dir = root_dir / "salmon" / "_out"
        for log in log_dir.glob("*.log"):
            _clear_logs(log=log)


def _reset(server):
    server.authorize()
    server.reset()

    # Delete files
    _clear_logs()
    OUT = Path(__file__).absolute().parent.parent / "salmon" / "_out"
    dump = OUT / "dump.rdb"
    if dump.exists():
        dump.unlink()
    assert not dump.exists()

    #  server.authorize()
    #  server.reset()
    return server


@pytest.fixture()
def server():
    server = Server("http://127.0.0.1:8421")
    server.authorize()
    # server = _reset(server)
    sleep(1)
    yield server
    server = _reset(server)
    sleep(5)


class LogError(Exception):
    pass


class Logs:
    def __init__(self):
        this_dir = Path(__file__).absolute().parent
        root_dir = this_dir.parent
        self.log_dir = root_dir / "salmon" / "_out"
        self.catch = True
        self.warn = True
        self.delay = 0.5  # for files to finish flushing

    def __enter__(self):
        _clear_logs()

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is not None:
            raise exc_type(exc_value)
        sleep(self.delay)
        files = list(self.log_dir.glob("*.log"))
        msg = f"files for checking logs = {files}"
        logger.warning(msg)
        _errors = []
        _warnings = []
        for log in files:
            lines = log.read_text().split("\n")
            for line in lines:
                if any(x in line.lower() for x in ["error", "except"]):
                    _errors.append(line)
                if "warn" in line.lower() and "answer received:" not in line.lower():
                    _warnings.append(line)
        if self.warn and len(_warnings):
            warn("\n".join(_warnings))
        if self.catch and len(_errors):
            raise LogError("\n".join(_errors))
        _clear_logs()


@pytest.fixture()
def logs():
    return Logs()


def alien_egg(head, left, right, random_state=None):
    """
    Parameters
    ---------- head, left, right : int, int, int
        Number of spikes on the various objects

    Returns
    -------
    winner : str
        Either "left" or "right"

    Notes
    -----
    This is determined from human data.
    See datasets/strange-fruit-triplet/noise-model.ipynb for details.
    """
    ldiff = np.abs(head - left)
    rdiff = np.abs(head - right)

    r = np.maximum(ldiff, rdiff) / (ldiff + rdiff)
    rate = 19.5269746
    final = 0.9567
    p_correct = final / (1 + np.exp(-rate * (r - 0.5)))

    winner = 0 if ldiff < rdiff else 1
    random_state = check_random_state(random_state)

    if random_state.uniform() <= p_correct:
        return winner
    return 1 - winner



import json
import os
import random
import threading
import traceback
from copy import deepcopy
from typing import Dict, Union

import cloudpickle
from dask.distributed import Client as DaskClient
from dask.distributed import fire_and_forget
from fastapi import BackgroundTasks, FastAPI, HTTPException
from fastapi.responses import PlainTextResponse
from rejson import Client, Path
from starlette.exceptions import HTTPException as StarletteHTTPException

from salmon.frontend.utils import ServerException
from salmon.triplets import samplers
from salmon.utils import flush_logger, get_logger

DEBUG = os.environ.get("SALMON_DEBUG", 0)

logger = get_logger(__name__)

root = Path.rootPath()
rj = Client(host="redis", port=6379, decode_responses=True)

app = FastAPI(title="salmon-backend")
threads = []

SAMPLERS = {}


def exception_to_string(excp):
    flush_logger(logger)
    stack = traceback.extract_stack() + traceback.extract_tb(
        excp.__traceback__
    )  # add limit=??
    pretty = traceback.format_list(stack)
    return "Error!\n\n\nSummary:\n\n{} {}\n\nFull traceback:\n\n".format(
        excp.__class__, excp
    ) + "".join(pretty)


class ExpParsingError(StarletteHTTPException):
    pass


@app.exception_handler(ExpParsingError)
async def http_exception_handler(request, exc):
    flush_logger(logger)
    return PlainTextResponse(exc.detail, status_code=exc.status_code)


@app.post("/init/{ident}")
async def init(ident: str, background_tasks: BackgroundTasks) -> bool:
    """
    Start running an algorithm.

    Parameters
    ----------
    ident : str
        The identifier paired with this algorithm.

    Returns
    -------
    success : bool

    Notes
    -----
    This function has side effects: it launches background job with
    algorithm class. This class runs the ``run`` function, which posts
    queries to Redis and process answers posted to Redis.

    If the algorithm class has a ``get_query`` method, the class will
    respond to the API request ``/get_query``. The method ``run`` should
    be modified to handle this.

    params : Dict[str, Any]
        Pulled from the experiment config and Redis.
        Here's an example YAML configuration:

    .. code:: yaml

       targets:
         - 1
         - 2
         - 3
         - 4
       samplers:
         - Random
         - random2
           - class: Random
           - foo: bar

    """
    # TODO: Better handling of exceptions if params keys don't match
    logger.info("backend: initializing %s", ident)
    config = rj.jsonget("exp_config")

    try:
        if f"state-{ident}" in rj.keys():
            logger.warning(f"Initializing alg from key 'state-{ident}'")
            # See https://github.com/andymccurdy/redis-py/issues/1006
            rj2 = Client(host="redis", port=6379, decode_responses=False)
            state = rj2.get(f"state-{ident}")
            alg = cloudpickle.loads(state)
        else:
            logger.warning(f"Initializing alg from config")
            params = config["samplers"][ident]
            _class = params.pop("class", ident)
            Sampler = getattr(samplers, _class)
            params = {k: _fmt_params(k, v) for k, v in params.items()}
            logger.warning("Sampler for %s = %s", ident, Sampler)
            common = config["sampling"]["common"]
            p = deepcopy(common)
            p.update(params)
            kwargs = dict(ident=ident, n=config["n"], **p)
            logger.warning("class=%s kwargs= %s", _class, kwargs)
            alg = Sampler(ident=ident, n=config["n"], **p)
    except Exception as e:
        msg = exception_to_string(e)
        logger.error(f"Error on alg={ident} init: {msg}")
        flush_logger(logger)
        raise ExpParsingError(status_code=500, detail=msg)

    SAMPLERS[ident] = alg

    dask_client = DaskClient("127.0.0.2:8786")
    logger.info("Before adding init task")
    background_tasks.add_task(alg.run, dask_client)
    logger.info("Returning")
    return True


@app.post("/reset/")
def reset():
    keys = deepcopy(list(SAMPLERS.keys()))
    for k in keys:
        SAMPLERS.pop(k)


@app.get("/query/{ident}")
def get_query(ident: str, puid: str=""):
    global SAMPLERS
    alg = SAMPLERS[ident]
    if hasattr(alg, "get_query"):
        try:
            kwargs = dict() if not puid else dict(puid=puid)
            q, score = alg.get_query(**kwargs)
            logger.debug("q, score = %s, %s", q, score)
        except Exception as e:
            logger.exception(e)
            flush_logger(logger)
            raise HTTPException(status_code=500, detail=str(e))
        if q is None:
            flush_logger(logger)
            raise HTTPException(status_code=404)
        return {"sampler": ident, "score": score, **q}


def _fmt_params(k, v):
    if isinstance(v, (str, int, float, bool, list)):
        return v
    elif isinstance(v, dict):
        return {f"{k}__{ki}": _fmt_params(ki, vi) for ki, vi in v.items()}
    raise ValueError(f"Error formatting key={k} with value {v}")


@app.get("/model/{sampler}")
async def get_model(sampler: str):
    samplers = rj.jsonget("samplers")
    if sampler not in samplers:
        raise ServerException(
            f"Can't find model for sampler='{sampler}'. "
            f"Valid choices for sampler are {samplers}"
        )
    if f"model-{sampler}" not in rj.keys():
        logger.warning("rj.keys() = %s", rj.keys())
        flush_logger(logger)
        raise ServerException(f"Model has not been created for sampler='{sampler}'")
    rj2 = Client(host="redis", port=6379, decode_responses=False)
    ir = rj2.get(f"model-{sampler}")
    model = cloudpickle.loads(ir)
    return model


@app.get("/meta/perf/{sampler}")
async def get_timings(sampler: str):
    samplers = rj.jsonget("samplers")
    if sampler not in samplers:
        msg = (
            f"Can't find key for sampler='{sampler}'. "
            f"Valid choices for sampler are {samplers}"
        )
        logger.warning(msg)
        raise ServerException(msg)
    keys = list(sorted(rj.keys()))
    if f"alg-perf-{sampler}" not in keys:
        logger.warning("rj.keys() = %s", keys)
        msg = (
            f"Performance data has not been created for sampler='{sampler}'. Database has keys {keys}"
        )
        logger.warning(msg)
        raise ServerException(msg)
    return rj.jsonget(f"alg-perf-{sampler}")



import numpy as np
import numpy.linalg as LA

from salmon.triplets.samplers.adaptive import UncertaintyScorer


def test_uncertainty_sampling(random_state=42):
    n, d = 100, 2
    rng = np.random.RandomState(random_state)
    X = rng.uniform(size=(n, d))

    search = UncertaintyScorer(embedding=X)
    queries, scores = search.score(num=100)
    distances = [
        abs(LA.norm(X[h] - X[l]) ** 2 - LA.norm(X[h] - X[r]) ** 2)
        for h, l, r in queries
    ]
    idx_best_score = np.argmax(scores)
    assert np.allclose(distances[idx_best_score], min(distances))



import pathlib
import random
from copy import copy
from datetime import datetime, timedelta
from textwrap import dedent
from time import time
from typing import Dict, Union

import numpy as np
import requests as httpx
import ujson
from fastapi import FastAPI
from redis.exceptions import ResponseError
from rejson import Client, Path
from starlette.requests import Request
from starlette.staticfiles import StaticFiles
from starlette.templating import Jinja2Templates
from starlette_exporter import PrometheusMiddleware, handle_metrics

from salmon.frontend.utils import ServerException, image_url, sha256
from salmon.triplets import manager
from salmon.utils import get_logger

logger = get_logger(__name__)

root = Path.rootPath()
rj = Client(host="redis", port=6379, decode_responses=True)


def start_algs():
    """
    Start the algorithm backend. This function is necessary because the
    machine might be restarted (so the experiment isn't launched fresh).
    """
    if "samplers" not in rj.keys():
        return
    names = rj.jsonget("samplers")
    for name in names:
        logger.info(f"Restarting alg={name}...")
        r = httpx.post(f"http://localhost:8400/init/{name}")
        assert r.status_code == 200
    return True


def stop_algs():
    rj.jsonset("reset", root, True)
    return True


app = FastAPI(
    title="Salmon",
    description=dedent(
        """A prototype platform for crowdsourcing triplet queries.
        \n\n***Warning!*** This platform is experimental and unstable.
        """
    ),
    on_startup=[start_algs],
    on_shutdown=[stop_algs],
)
app.add_middleware(PrometheusMiddleware)
app.add_route("/metrics", handle_metrics)

pkg_dir = pathlib.Path(__file__).absolute().parent
app.mount("/static", StaticFiles(directory=str(pkg_dir / "static")), name="static")
templates = Jinja2Templates(directory="templates")


async def _get_config():
    return rj.jsonget("exp_config")


async def _ensure_initialized():
    if "exp_config" not in rj:
        raise ServerException("No data has been uploaded")
    exp_config = await _get_config()
    expected_keys = ["targets", "samplers", "n", "sampling", "html"]
    html_keys = [
        "instructions",
        "max_queries",
        "debrief",
        "skip_button",
        "css",
        "arrow_keys",
    ]
    err = False
    if not set(exp_config) == set(expected_keys):
        err = True
        extra = set(exp_config) - set(expected_keys)
        missing = set(expected_keys) - set(exp_config)
    if "html" in exp_config and not set(html_keys).issubset(set(exp_config["html"])):
        err = True
        extra = set()  # exp_config["html"]) - set(expected_keys)
        missing = set(expected_keys) - set(exp_config["html"])
    if err:
        msg = (
            "Experiment keys are not correct. Expected {}, got {}.\n\n"
            "Extra keys: {}\n"
            "Missing keys: {}"
        )
        raise ServerException(
            msg.format(expected_keys, list(exp_config.keys()), extra, missing)
        )
    return exp_config


@app.get("/", tags=["public"])
async def get_query_page(request: Request, puid: str = ""):
    """
    Load the query page and present a "triplet query".
    """
    exp_config = await _ensure_initialized()
    if puid == "":
        uid = "salmon-{}".format(np.random.randint(2 ** 32 - 1))
        puid = sha256(uid)[:16]
    try:
        urls = [image_url(t) for t in exp_config["targets"]]
    except:
        urls = []
    items = {
        "puid": puid,
        "targets": exp_config["targets"],
        "samplers_per_user": exp_config["sampling"]["samplers_per_user"],
        "urls": urls,
        "details": exp_config["sampling"]["details"],
        "html": exp_config["html"],
    }
    items.update(request=request)
    return templates.TemplateResponse("query_page.html", items)


@app.get("/query", tags=["public"])
async def get_query(sampler="", puid="") -> Dict[str, Union[int, str, float]]:
    if sampler == "":
        samplers = rj.jsonget("samplers")
        probs = rj.jsonget("sampling_probs")

        idx = np.random.choice(len(samplers), p=probs)
        sampler = samplers[idx]

    host = f"http://localhost:8400"
    endpoint = f"/query/{sampler}"
    if puid:
        endpoint = endpoint + f"?puid={puid}"

    r = httpx.get(host + endpoint)
    if r.status_code == 200:
        logger.info(f"query r={r}")
        return r.json()

    key = f"alg-{sampler}-queries"
    logger.info(f"zpopmax'ing {key}")
    queries = rj.zpopmax(key)
    if len(queries):
        serialized_query, score = queries[0]
        q = manager.deserialize_query(serialized_query)
    else:
        config = await _get_config()
        q = manager.random_query(config["n"])
        score = -9999

    return {"sampler": sampler, "score": score, **q}


@app.post("/answer", tags=["public"])
async def process_answer(ans: manager.Answer):
    """
    Process the answer, and append the received answer (alongside participant
    UID) to the database.

    See the <a href='#model-Answer'>Answer schema</a> for more detail.

    Returns
    -------
    `d : Dict[str, bool]`. On success, `d == {"success": True}`

    """
    d = ujson.loads(ans.json())
    _update = {
        "time_received": round(time(), 3),
        "loser": d["left"] if d["winner"] == d["right"] else d["right"],
    }
    d.update(_update)
    ident = d["sampler"]
    logger.warning(f"answer received: {d}")
    rj.jsonarrappend(f"alg-{ident}-answers", root, d)
    # on backend,  key = f"alg-{self.ident}-answers"
    rj.jsonarrappend("all-responses", root, d)
    last_save = rj.lastsave()

    # Save every 15 minutes
    if (datetime.now() - last_save) >= timedelta(seconds=60 * 15):
        try:
            rj.bgsave()
        except ResponseError as e:
            if "Background save already in progress" not in str(e):
                raise e

    return {"success": True}



import logging
import pickle
from typing import Any, Dict, List, Tuple

from pydantic import BaseModel

from salmon.utils import get_logger

Query = Tuple[int, Tuple[int, int]]  # head, (choice 1, choice 2)

logger = get_logger(__name__)


class Answer(BaseModel):
    head: int
    left: int
    right: int
    winner: int



import asyncio
import hashlib
import itertools
import json
import os
import pathlib
import pprint
import shutil
import sys
import traceback
from copy import deepcopy
from datetime import datetime, timedelta
from io import StringIO
from textwrap import dedent
from time import sleep, time
from typing import Any, Dict, List, Optional

import numpy as np
import pandas as pd
import requests as httpx
import yaml
from bokeh.embed import json_item
from fastapi import Depends, File, Form, HTTPException
from fastapi.responses import (FileResponse, HTMLResponse, JSONResponse,
                               PlainTextResponse)
from fastapi.security import HTTPBasic, HTTPBasicCredentials
from redis import ResponseError
from rejson import Client, Path
from starlette.exceptions import HTTPException as StarletteHTTPException
from starlette.requests import Request
from starlette.status import HTTP_401_UNAUTHORIZED, HTTP_409_CONFLICT

import salmon
from salmon.frontend import plotting
from salmon.frontend.public import _ensure_initialized, app, templates
from salmon.frontend.utils import (ServerException, _extract_zipfile,
                                   _format_target, _format_targets, get_logger)
from salmon.triplets import manager
from salmon.triplets.manager import Config

security = HTTPBasic()

root = Path.rootPath()
rj = Client(host="redis", port=6379, decode_responses=True)
logger = get_logger(__name__)
DIR = pathlib.Path(__file__).absolute().parent
ROOT_DIR = DIR.parent.parent
OUT_DIR = DIR.parent / "_out"
assert OUT_DIR.exists(), str(OUT_DIR)

CREDS_FILE = OUT_DIR / "creds.json"
EXPECTED_PWORD = "331a5156c7f0a529ed1de8d9aba35da95655c341df0ca0bbb2b69b3be319ecf0"


def _salt(password: str) -> str:
    pword = bytes(password, "utf8")
    salt = b"\x87\xa4\xb0\xc6k\xb7\xcf!\x8a\xc8z\xc6Q\x8b_\x00i\xc4\xbd\x01\x15\xabjn\xda\x07ZN}\xfd\xe1\x0e"
    m = hashlib.sha256()
    m.update(pword + salt)
    return m.digest().hex()


def _write_user_pass(user: str, password: str) -> bool:
    if not CREDS_FILE.exists():
        passwords = {}
    else:
        with open(CREDS_FILE, "r") as f:
            passwords = json.load(f)

    logger.warning(f"passwords={passwords}")
    if user in passwords:
        raise HTTPException(
            status_code=403, detail=f"user={user} already in password file.",
        )
    new_user = user not in passwords
    if new_user and len(passwords) > 0:
        raise HTTPException(
            status_code=403,
            detail="A maximum number of users have already been registered",
        )

    passwords[user] = _salt(password)
    with open(CREDS_FILE, "w") as f:
        json.dump(passwords, f)

    return True


@app.post("/create_user/{username}/{password}", tags=["private"])
def _create_user_v0(username: str, password: str):
    _write_user_pass(username, password)


@app.post("/create_user", tags=["private"])
def create_user(
    username: str = Form(...), password: str = Form(...), password2: str = Form(...)
) -> bool:
    logger.warning(f"In create_user with username={username}, password={password}")
    if password != password2:
        raise HTTPException(
            status_code=HTTP_409_CONFLICT,
            detail=f"password={password} does not match the confirmation password={password2}",
            headers={"WWW-Authenticate": "Basic"},
        )
    _write_user_pass(username, password)
    msg = dedent(
        f"""
        <p>
          Success! User=<code>{username}</code> successfully created with
          password=<code>{password}</code>.
        </p>
        <p>
          <b>Do NOT lose this information.</b> Salmon will be <b>unusable</b>
          without this, and this is the last time this information will be
          displayed. Tasks that require this information include downloading
          the responses collected by Salmon and viewing the dashboard.
        </p>
        <p>
          <b>Salmon will only accept this username/password.</b>
        </p>
        <p>
          <b>Next,</b> visit <a href='/init'>/init</a> to initialize a new
          experiment with this username={username} and password={password}.
        </p>
        """
    )
    return HTMLResponse(msg)


def _authorize(creds: HTTPBasicCredentials = Depends(security)) -> bool:
    if CREDS_FILE.exists():
        with open(CREDS_FILE, "r") as f:
            passwords = json.load(f)
    else:
        passwords = {}

    name = creds.username
    if (name in passwords and _salt(creds.password) == passwords[name]) or (
        name == "foo8421" and _salt(creds.password) == EXPECTED_PWORD
    ):
        logger.info("Authorized: true")
        return True

    logger.info("Not authorized!")
    raise HTTPException(
        status_code=HTTP_401_UNAUTHORIZED,
        detail="Incorrect username or password",
        headers={"WWW-Authenticate": "Basic"},
    )


@app.get("/init", tags=["private"])
def upload_form():
    """
    Upload a YAML file that specifies an experiment. See
    <a href='#operations-private-process_form_init_exp_post'>
    the POST description</a> for more detail on the file.

    """
    warning = ""
    if rj.jsonget("exp_config"):
        warning = dedent(
            """<div style="color: #f00;">
            <p>Warning: an experiment is already set!
               Visit [url]:8421/reset to reset the expeirment</p>
            </div>
            """
        )

    password = dedent(
        """
        <h2 style="text-align: center;">Step 1: create new username/password.</h2>
        <div style="text-align: center; padding: 10px;">
        <p>This username/password distinguishes you from any other internet
        user.</p>
        <form action="/create_user" enctype="multipart/form-data" method="post">
        <ul>
          <li>Username: <input name="username" placeholder="username"
                         type="text" /></li>
          <li>Password: <input name="password" placeholder="password"
                         type="text" /></li>
          <li>Confirm password: <input name="password2" placeholder="password"
                                 type="text" /></li>
        </ul>
        <input type="submit" value="Create user">
        </form>
        <p>
          <b>Do not lose this username/password!</b>
          After you've created a username/password <i>and written it down
          </i>, then you can create a new experiment. It'll ask for the
          username/password you just created.</p>
        </div>
        """
    )
    if CREDS_FILE.exists():
        with open(CREDS_FILE, "r") as f:
            passwords = json.load(f)
        logger.warning(f"passwords={passwords}")
        if len(passwords):
            user = list(passwords.keys())[0]
            letters = list(user)
            for k, letter in enumerate(letters):
                if k in [0, len(letters) - 1]:
                    continue
                letters[k] = "*"
            user = "".join(letters)
            password = (
                "<div style='text-align: center; padding: 10px;'>"
                f"<p>A user with name <code>{user}</code> has been created "
                "(only first and last letters shown). "
                "Do you know the password?</p>"
                "</div>"
            )
    body = dedent(
        f"""<body>
        <div style="display: table; margin: 0 auto; max-width: 600px;">
        <div>
        <ul>
          <li>
            <a href="https://stsievert.com/salmon/">
              General Salmon documentation
            </a>
          </li>
          <li>
            <a href="/docs#/private/process_form_init_exp_post">
              Detailed /init_exp endpoint documentation
            </a>
          </li>
        </div>
        <div style="text-align: center; padding: 10px;">
        <p>Two steps are required:<p>
        <ol>
        <li>Creating a username/password.</li>
        <li>Creating a new experiment.</li>
        </ol>
        </div>
        {password}
        <h2 style="text-align: center;">Step 2: create new experiment.</h2>
        <h3 style="text-align: center;">Option 1: initialize new experiment.</h3>
        <div style="text-align: center; padding: 10px;">
        <form action="/init_exp" enctype="multipart/form-data" method="post">
        <ul>
          <li>Experiment parameters (YAML file): <input name="exp" type="file"></li>
          <li>Images/movies (ZIP file, optional): <input name="targets" type="file"></li>
        </ul>
        <input type="submit" value="Create experiment">
        </form>
        {warning}
        </div>
        <h3 style="text-align: center;">Option 2: restore an old experiment.</h3>
        <div style="text-align: center; padding: 10px;">
        <p>Instructions:
        <ol style="text-align: left;">
        <li>Upload database dump from Salmon. The name should look like
          <code>exp-2020-03-12-v0.3.1.rdb</code> if downloaded on March 12th, 2020.</li>
        <li>Restart the server. On Amazon EC2, this means choosing the EC2 instance state "reboot".</li>
        </ol>
        </p>
        <form action="/init_exp" enctype="multipart/form-data" method="post">
          <ul>
            <li>Database file : <input name="rdb" type="file"></li>
          </ul>
          <input type="submit" value="Create experiment">
        </form>
        <p>
          This may fail. See the documentation FAQs for
          more detail and tips to resolve it:
          <a href="https://docs.stsievert.com/salmon/installation#troubleshooting">https://docs.stsievert.com/salmon/installation#troubleshooting</a>.
        </p>
        </div>
        </div>
        </body>
        """
    )
    return HTMLResponse(content=body)


@app.get("/config", tags=["private"])
async def _get_config_endpoint(json: bool = True):
    exp_config = await _ensure_initialized()
    exp_config.pop("n")
    print("json=", json, bool(json), not json)
    if not json:
        return PlainTextResponse(yaml.dump(exp_config))
    return JSONResponse(exp_config)


async def _get_config(exp: bytes, targets: bytes) -> Dict[str, Any]:
    user_config = yaml.safe_load(exp)
    logger.warning(f"exp = {exp}")
    logger.warning(f"user_config = {user_config}")

    if targets:
        fnames = _extract_zipfile(targets)
        logger.info("fnames = %s", fnames)
        if len(fnames) == 1 and ".csv" in fnames[0].suffixes:
            T = _format_targets(fnames[0])
        else:
            T = [_format_target(f) for f in fnames]
    elif isinstance(user_config["targets"], int):
        T = [str(x) for x in range(user_config["targets"])]
    else:
        T = [str(x) for x in user_config["targets"]]
    user_config["targets"] = T

    c = Config()  # defaults already encoded
    c = c.parse(user_config)
    config = c.dict()

    logger.warning(config["sampling"]["probs"])
    logger.warning(config["samplers"])

    config["n"] = len(config["targets"])
    logger.info("initializing experinment with %s", config)
    return config


def exception_to_string(excp):
    stack = traceback.extract_stack() + traceback.extract_tb(excp.__traceback__)
    pretty = traceback.format_list(stack)
    return "Error:\n\n{}\n\nMessage:\n\n{}\n\n\nSummary:\n\n{} {}\n\nTraceback:\n\n".format(
        str(excp), getattr(excp, "detail", ""), excp.__class__, excp
    ) + "".join(
        pretty
    )


class ExpParsingError(StarletteHTTPException):
    pass


@app.exception_handler(ExpParsingError)
async def http_exception_handler(request, exc):
    return PlainTextResponse(exc.detail, status_code=exc.status_code)


@app.post("/init_exp", tags=["private"])
async def process_form(
    request: Request,
    exp: bytes = File(default=""),
    targets: bytes = File(default=""),
    rdb: bytes = File(default=""),
    authorized: bool = Depends(_authorize),
):
    """
    The uploaded file needs to have the following keys:

    * targets (List)
    * instructions (Optional[str])
    * max_queries (Optional[int])

    Targets/instructions can render most HTML tags.

    Example
    -------

        - targets:
          - object 1
          - object 2
          - <b>bold</i> object 3
          - <i>object</i> 4
          - <img src="https://en.wikipedia.org/wiki/File:2010_Winter_Olympics_Bode_Miller_in_downhill.jpg" />
        - instructions: "Foobar!"
        - max_queries: 25
    """
    try:
        if rj.jsonget("exp_config"):
            detail = (
                "Incorrect username or password",
                "An experiment is already set! This experiment has not been "
                "deleted, and a new experiment has not been initialized."
                "\n\nIt is possible to clear this message; however, "
                "care needs to be taken to ensure you want to initialize a new"
                "experiment. To clear this error, follow these steps"
                "\n\n1. Verify which experiment is running. Uploading a new"
                "experiment will overwrite this experiment. Do you mean to upload?"
                "\n2. Visit /reset. Warning: this will *delete* the experiment"
                "\n3. Revisit /init and re-upload the experiment."
                "\n\n(visiting /foo means visiting '[url]:8421/foo'",
            )
            experiment_set = True
            raise HTTPException(status_code=403, detail=detail)
        experiment_set = False
        ret = await _process_form(request, exp, targets, rdb)

        if rdb:
            return ret

        await _ensure_initialized()
        return ret
    except Exception as e:
        logger.error(e)
        if authorized and not experiment_set:
            _reset(timeout=5)
        if isinstance(e, (ExpParsingError, HTTPException)):
            raise e
        msg = exception_to_string(e)
        logger.error(msg)
        raise ExpParsingError(status_code=500, detail=msg)


async def _process_form(
    request: Request,
    exp: bytes = File(default=""),
    targets: bytes = File(default=""),
    rdb: bytes = File(default=""),
):
    if rdb:
        return await restore(rdb=rdb)
    logger.info("salmon.__version__ = %s", app.version)
    exp_config = await _get_config(exp, targets)

    rj.jsonset("exp_config", root, exp_config)

    # Start the backend
    names = list(exp_config["samplers"].keys())
    _probs = exp_config["sampling"]["probs"]
    probs = [_probs[n] / 100 for n in names]
    rj.jsonset("samplers", root, names)
    rj.jsonset("sampling_probs", root, probs)
    for name in names:
        rj.jsonset(f"alg-{name}-answers", root, [])

        # Don't touch! Not set because rj.zadd doesn't require it.
        # rj.jsonset(f"alg-{name}-queries", root, [])

        logger.info(f"initializing algorithm {name}...")
        r = httpx.post(f"http://localhost:8400/init/{name}")
        if r.status_code != 200:
            msg = "Algorithm errored on initialization.\n\n" + r.text
            logger.error("Error! r.text = %s", r.text)
            logger.error(msg)
            raise ExpParsingError(status_code=500, detail=msg)
        logger.info(f"done initializing {name}.")

    _time = time()
    rj.jsonset("start_time", root, _time)
    rj.jsonset("start_datetime", root, datetime.now().isoformat())
    rj.jsonset("all-responses", root, [])
    _save(rj)

    nice_config = pprint.pformat(exp_config)
    logger.info("Experiment initialized with\nexp_config=%s", nice_config)
    response = dedent(
        """<html><body>
        <br><br>
        <p>
        Now, Salmon presents the following interfaces:
        </p>
        <p><ul style="text-align: center;">
        <li><a href="/">Query page</a>. Send this page to crowdsourcing participants.</li>
        <li><a href="/dashboard">Dashboard</a>. Use this page to monitor experimental progress.</li>
        </ul></p>
        </body></html>
        """
    )
    return HTMLResponse(content=response)


@app.delete("/reset", tags=["private"])
def reset_delete(
    force: int = 0,
    authorized=Depends(_authorize),
    tags=["private"],
    timeout: float = 10,
):
    if authorized:
        reset(force=force, authorized=authorized, timeout=timeout)
        return {"success": True}
    return {"success": False}


@app.get("/reset", tags=["private"])
def reset(
    force: int = 0,
    authorized=Depends(_authorize),
    tags=["private"],
    timeout: float = 10,
):
    """Reset the machine to it's initial state.

    This deletes all data from the database (and clear the backend too).
    It also removes the username/password.
    """
    if not authorized:
        return {"success": False}

    logger.warning("Resetting, force=%s, authorized=%s", force, authorized)
    if not force:
        logger.warning("Resetting, force=False. Erroring")
        msg = """Do you really want to delete *all* data? This will delete all
            responses and all target information and *cannot be undone.*

            If you do really want to reset, go to '[url]/reset?force=1'
            instead of '[url]/reset'. That means these actions will occur:

            * Any experiment data will be deleted. Responses, embeddings, etc.
            * Your username/password will be deleted.

            These actions are final and can not be undone.
            """
        raise ServerException(dedent(msg), status_code=403)

    logger.warning("Authorized reset, force=True. Removing data from database")
    meta = _reset(timeout=timeout)
    assert meta["success"]

    now = datetime.now().isoformat()[:16]
    logger.warning(
        f"Authorized reset, force=True. Removing creds.json to creds-{now}.json"
    )
    CREDS_FILE.rename(OUT_DIR / f"creds-{now}.json")
    assert not CREDS_FILE.exists()

    return HTMLResponse("This Salmon server has been reset to it's initial state.")


def _save(rj, bg=False):
    try:
        if bg:
            rj.bgsave()
        else:
            rj.save()
    except ResponseError as e:
        if "save already in progress" not in str(e):
            raise e


def _reset(timeout: float = 5):
    _save(rj)

    # Stop background jobs (ie adaptive algs)
    rj.jsonset("reset", root, True)
    rj2 = Client(host="redis", port=6379, decode_responses=False)
    if "samplers" in rj.keys():
        samplers = rj.jsonget("samplers")
        stopped = {name: False for name in samplers}
        __deadline = time() + timeout
        for k in itertools.count():
            rj.jsonset("reset", root, True)
            for name in stopped:
                if f"stopped-{name}" in rj.keys():
                    stopped[name] = rj.jsonget(f"stopped-{name}", root)
            if all(stopped.values()):
                logger.warning(f"stopped={stopped}")
                break
            sleep(1)
            logger.warning(
                f"Waited {k + 1} seconds algorithms... stopped? {stopped}"
                f" (rj.keys() == {rj.keys()}"
            )
            if timeout and time() >= __deadline:
                logger.warning(f"Hit timeout={timeout} w/ stopped={stopped}. Breaking!")
                break

        logger.warning("    starting with clearing queries...")
        for ident in samplers:
            rj2.delete(f"alg-{ident}-queries")
    httpx.post(f"http://localhost:8400/reset/")

    logger.warning("Trying to completely flush database...")

    for _rj in [rj, rj2]:
        _rj.memory_purge()
        _rj.flushall(asynchronous=True)
        _rj.flushdb(asynchronous=True)
        _rj.memory_purge()
        for k in _rj.keys():
            _rj.delete(k)
        _rj.flushdb(asynchronous=False)
        _rj.flushall(asynchronous=False)

    now = datetime.now().isoformat()[: 10 + 6]

    save_dir = OUT_DIR
    files = [f.name for f in save_dir.glob("*")]
    logger.warning(f"save_dir = {str(save_dir)}")
    logger.warning(f"dump_rdb in files? {'dump.rdb' in files}")
    if "dump.rdb" in files:
        logger.warning(f"Moving dump.rdb to dump-{now}.rdb")
        shutil.move(str(save_dir / "dump.rdb"), str(save_dir / f"dump-{now}.rdb"))
        files = [f.name for f in save_dir.glob("*")]
        logger.warning(f"after moving, dump_rdb in files? {'dump.rdb' in files}")
    files = [f.name for f in save_dir.glob("*")]
    assert "dump.rdb" not in files

    logger.warning("After reset, rj.keys=%s", rj.keys())
    rj.jsonset("responses", root, {})
    rj.jsonset("start_time", root, -1)
    rj.jsonset("start_datetime", root, "-1")
    rj.jsonset("exp_config", root, {})
    return {"success": True}


@app.get("/responses", tags=["private"])
async def get_responses(
    authorized: bool = Depends(_authorize), json: Optional[bool] = True
) -> Dict[str, Any]:
    """
    Get the recorded responses from the current experiments. This includes
    the following columns:

    * `left`, `right`, `head`: Indices describing the objects in the
      head/left/right positions.
    * `head_html`, `right_html`, `left_html`: the HTML
      representation of the target in the head/left/right position.
    * `datetime_received`: the time the response was received.
    * `response_time`: the time spent between the providing the query and
      receiving the answer.

    There may be additional columns.

    Returns
    -------
    The list of responses as a CSV file. This file can be read by
    Panda's `read_csv` function.

    """
    exp_config = await _ensure_initialized()
    targets = exp_config["targets"]
    start = rj.jsonget("start_time")
    responses = await _get_responses()
    json_responses = await _format_responses(responses, targets, start)

    if json:
        return JSONResponse(
            json_responses,
            headers={"Content-Disposition": 'attachment; filename="responses.json"'},
        )
    with StringIO() as f:
        df = pd.DataFrame(json_responses)
        df.to_csv(f, index=False)
        out = f.getvalue()

    return PlainTextResponse(
        out, headers={"Content-Disposition": 'attachment; filename="responses.csv"'}
    )


def _fmt_embedding(
    embedding: List[List[float]], targets: List[str], **kwargs
) -> pd.DataFrame:
    df = pd.DataFrame({"target_html": targets})
    df["target_id"] = np.arange(len(df)).astype(int)
    for k, v in kwargs.items():
        df[k] = v

    em = np.asarray(embedding)
    if em.ndim == 1:
        em = em.reshape(1, -1)
    for k2, col in enumerate(range(em.shape[1])):
        df[k2] = em[:, col]
    return df


@app.get("/embeddings", tags=["private"])
async def get_embeddings(
    authorized: bool = Depends(_authorize), sampler: Optional[str] = None,
):
    """
    Get the embeddings for algorithms.

    Parameters
    ----------

    * sampler : str, optional. The algorithm to get the embedding for.

    Returns
    -------
    CSV with columns for the target HTML, target ID, the embedding, and the
    algorithm generating the embedding.
    """
    exp_config = await _ensure_initialized()
    exp_config = deepcopy(exp_config)
    targets = exp_config.pop("targets")
    samplers = list(exp_config.pop("samplers").keys())
    embeddings = {}
    for s in samplers:
        try:
            embeddings[s] = await get_model(s)
        except:
            pass
    if len(embeddings) == 0:
        raise ServerException(
            f"No model has been created for any sampler in {samplers}"
        )
    dfs = {
        alg: _fmt_embedding(model["embedding"], targets, alg=alg)
        for alg, model in embeddings.items()
    }

    if sampler is not None:
        df = dfs[sampler]
    else:
        df = pd.concat(dfs)

    with StringIO() as f:
        df.to_csv(f, index=False)
        out = f.getvalue()

    fname = "embeddings.csv" if sampler is None else f"embedding-{sampler}.csv"
    return PlainTextResponse(
        out, headers={"Content-Disposition": f'attachment; filename="{fname}"'}
    )


async def _get_responses():
    """
    Get the recorded responses. This JSON file is readable by Pandas:
    <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.read_json.html>

    Returns
    -------
    `json_file : str`. This file will have keys

    * `head`, `left`, `right`, `winner` as integers describing the arms
      (and `_html`/`_src` as their HTML string/HTML `src` tag)
    * `puid` as the participant unique ID
    * `time_received_since_start`, an integer describing the time in
      seconds since launch start
    * `time_received`: the time in seconds since Jan. 1st, 1970.

    This file will be downloaded.

    """
    responses = rj.jsonget("all-responses", root)
    return responses


async def _format_responses(responses, targets, start):
    logger.info("getting %s responses", len(responses or []))
    out = manager.get_responses(responses, targets, start_time=start)
    return out


@app.get("/dashboard", tags=["private"])
@app.post("/dashboard", tags=["private"])
async def get_dashboard(request: Request, authorized: bool = Depends(_authorize)):
    """
    The primary method to view information about the experiment.
    It displays the following information:

    * Basic meta information: how many responses have been received, when
      were they received, etc.
    * Links to other API endpoints. These endpoint allow experiment
      download, getting the resposnes, resetting the experiment, etc.
    * Relevant graphs. Some answered questions are "how long did
      participants take to respond?" and "how long did it take to serve a
      query page?"
    """
    logger.info("Getting dashboard")
    _save(rj, bg=True)
    exp_config = await _ensure_initialized()
    exp_config = deepcopy(exp_config)
    targets = exp_config.pop("targets")
    start = rj.jsonget("start_time")
    start_datetime = timedelta(seconds=start) + datetime(1970, 1, 1)

    responses = await _get_responses()
    df = pd.DataFrame(
        responses, columns=["puid", "time_received", "response_time", "network_latency"]
    )
    df["start_time"] = start

    try:
        activity = await plotting.activity(df, start)
        response_times = await plotting.response_time(df)
        network_latency = await plotting.network_latency(df)
        response_rate = await plotting.response_rate(df)
        plots = {
            "activity": activity,
            "response_times": response_times,
            "network_latency": network_latency,
            "response_rate": response_rate,
        }
        plots = {k: json.dumps(json_item(v)) for k, v in plots.items()}
    except Exception as e:
        logger.exception(e)
        activity = response_times = network_latency = f"Exception! {e}"
        plots = {
            "activity": activity,
            "response_times": response_times,
            "network_latency": network_latency,
        }

    try:
        x = df["time_received"]
        rr_cdf, gaps_hist, response_meta = await plotting._get_response_rate_plots(x)
        plots["response_rate_cdf"] = json.dumps(json_item(rr_cdf))
        plots["gaps_histogram"] = json.dumps(json_item(gaps_hist))
    except Exception as e:
        logger.exception(e)
        plots["response_rate_cdf"] = {"/": "exception"}
        plots["gaps_histogram"] = {"/": "exception"}
        response_meta = {}

    try:
        endpoint_timing = await plotting.get_endpoint_time_plots()
        plots["endpoint_timings"] = {
            k: json.dumps(json_item(v)) for k, v in endpoint_timing.items()
        }
    except Exception as e:
        logger.exception(e)
        endpoint_timing = {"/": "exception"}
        plots["endpoint_timings"] = endpoint_timing

    endpoints = list(reversed(sorted(endpoint_timing.keys())))

    idents = rj.jsonget("samplers")
    models = {}
    for alg in idents:
        try:
            models[alg] = await get_model(alg)
        except Exception as e:
            models[alg] = None
    alg_plots = {
        alg: await plotting.show_embedding(model["embedding"], targets, alg=alg)
        for alg, model in models.items()
        if model
    }
    alg_plots = {k: json.dumps(json_item(v)) for k, v in alg_plots.items() if v}
    if not len(alg_plots):
        alg_plots = {"no embeddings": None}
    logger.info(f"idents = {idents}")

    perfs = {}
    for ident in idents:
        try:
            perfs[ident] = await _get_alg_perf(ident)
        except Exception as e:
            logger.exception(e)
            perfs[ident] = None

    _alg_perfs = {}
    for alg, data in perfs.items():
        if data:
            try:
                _alg_perfs[alg] = await plotting._get_alg_perf(pd.DataFrame(data))
            except Exception as e:
                logger.exception(e)
                _alg_perfs[ident] = "Error getting performace"
    alg_perfs = {
        k: json.dumps(json_item(v)) if not isinstance(v, str) else v
        for k, v in _alg_perfs.items()
    }
    if not len(alg_perfs):
        alg_perfs = {"no sampler timings": "none"}

    try:
        _query_db = {
            alg: await plotting._get_query_db(pd.DataFrame(data))
            for alg, data in perfs.items()
            if data
        }
        query_db = {k: json.dumps(json_item(v)) for k, v in _query_db.items() if v}
    except Exception as e:
        logger.exception(e)
        query_db = {"/": "Error getting query database stats"}
    if not len(query_db):
        query_db = {"no queries in database": "none"}

    return templates.TemplateResponse(
        "dashboard.html",
        {
            "start": start_datetime.isoformat()[: 10 + 6],
            "request": request,
            "targets": targets,
            "exp_config": exp_config,
            "num_responses": len(responses or []),
            "num_participants": df.puid.nunique(),
            "filenames": [_get_filename(html) for html in targets],
            "endpoints": endpoints,
            "alg_models": models,
            "alg_model_plots": alg_plots,
            "alg_perfs": alg_perfs,
            "config": exp_config,
            "samplers": idents,
            "query_db_perfs": query_db,
            **plots,
            **response_meta,
        },
    )


@app.get("/logs", tags=["private"])
async def get_logs(request: Request, authorized: bool = Depends(_authorize)):
    """
    Get detailed information about the server. This might include detailed
    tracebacks and exceptions.

    Returns
    -------
    JSON response with structure Dict[str, List[str]]

    The keys are the names of the different loggers, and each element in
    the list is one log record.
    """
    logger.info("Getting logs")

    items = {"request": request}
    files = OUT_DIR.glob("*.log")
    out = {}
    for file in files:
        with open(str(file), "r") as f:
            out[file.name] = f.readlines()
    return JSONResponse(out)


@app.get("/meta", tags=["private"])
async def get_meta(request: Request, authorized: bool = Depends(_authorize)):
    """
    Get meta information about the experiment.
    How many responses and participants are there?

    Returns
    -------
    JSON response describing meta information.
    """
    responses = await _get_responses()
    df = pd.DataFrame(responses, columns=["puid"])
    out = {
        "responses": len(df),
        "participants": df.puid.nunique(),
    }
    return JSONResponse(out)


@app.get("/download", tags=["private"])
async def download(request: Request, authorized: bool = Depends(_authorize)):
    """
    Download any experiment state. Functionally, this endpoint allows
    moving the experiment to a new machine.

    Returns
    -------
    experiment_state : file

    This file can be used to restore the contents of the Redis
    database on a new machine.
    """
    _save(rj, bg=False)
    fname = datetime.now().isoformat()[: 10 + 6]
    version = salmon.__version__
    headers = {
        "Content-Disposition": f'attachment; filename="exp-{fname}-salmon-{version}.rdb"'
    }
    return FileResponse(str(OUT_DIR / "dump.rdb"), headers=headers)


@app.post("/restore", tags=["private"])
async def restore(
    rdb: bytes = File(default=""), authorized: bool = Depends(_authorize)
):
    """
    Restore an experiment. This endpoint takes an experiment file from
    `/download` and restores it on the current machine.

    An experiment can be restored with the following steps:

    1. Download the experiment state at ``/download``
    2. Save the file on your machine.
    3. On a new machine, upload the file at ``/restore`` (this endpoint)
    4. Restart the machine, either via `docker-compose down; docker-
       compose up` or "Actions > Instance state > Reboot" on Amazon EC2.

    """
    with open(str(OUT_DIR / "dump.rdb"), "wb") as f:
        f.write(rdb)
    msg = dedent(
        """
        <div style="display: table; margin: 0 auto; max-width: 600px;">
        <br><br>
        <p><i><b>Your experiment is not initialized yet! Restart is required to restore experiment.</i></b></p>
        <p>
        To do this on Amazon EC2, select the \"Actions > Instance State > Reboot\"
        </p>
        <p>For developers, a "restart" means <code>docker-compose stop; docker-compose start</code>.</p>
        <p>After you reboot, visit the dashboard at
        <code>[url]:8421/dashboard</code>.
        <b>Do not visit the dashboard now</b>.</p>
        </div>
        """
    )
    return HTMLResponse(msg)


@app.get("/model/{sampler}")
async def get_model(sampler: str) -> Dict[str, Any]:
    logger.info("In public get_model with rj.keys() == %s", rj.keys())
    r = httpx.get(f"http://localhost:8400/model/{sampler}")
    if r.status_code != 200:
        raise ServerException(r.json()["detail"])
    return r.json()


async def _get_alg_perf(sampler: str) -> Dict[str, Any]:
    logger.info("In private _get_alg_perf with rj.keys() == %s", rj.keys())
    r = httpx.get(f"http://localhost:8400/meta/perf/{sampler}")
    if r.status_code != 200:
        raise ServerException(r.json()["detail"])
    return r.json()


def _get_filename(html):
    html = str(html)
    if "<img" in html or "<video" in html:
        i = html.find("src=")
        j = html[i:].find(" ")
        return html[i + 5 : i + j - 1].replace("/static/targets/", "")
    return html



import itertools
from collections import defaultdict
from copy import deepcopy
from textwrap import dedent
from time import time
from typing import Any, Dict, List, Optional, Tuple, TypeVar

import numpy as np
import numpy.linalg as LA
import pandas as pd
import torch.optim

import salmon.triplets.samplers.adaptive as adaptive
from salmon.backend.sampler import Sampler
from salmon.triplets.samplers._random_sampling import \
    _get_query as _random_query
from salmon.triplets.samplers.adaptive import InfoGainScorer, UncertaintyScorer
from salmon.utils import get_logger

logger = get_logger(__name__)

Query = TypeVar("Query")
Answer = TypeVar("Answer")


class Adaptive(Sampler):
    """
    The sampler that runs adaptive algorithms.
    """

    def __init__(
        self,
        *,
        n: int,
        d: int = 2,
        ident: str = "",
        module: str = "TSTE",
        optimizer: str = "Embedding",
        R: float = 10,
        scorer: str = "infogain",
        random_state: Optional[int] = None,
        **kwargs,
    ):
        """
        Parameters
        ----------
        n : int
            The number of items to embed.
        d : int (optional, default: ``2``)
            Embedding dimension.
        ident : str (optional, default: ``""``).
            The identity of this runner. Must be unique among all adaptive algorithms.
        optimizer : str (optional, default: ``Embedding``).
            The optimizer underlying the embedding. This method specifies how to
            change the batch size. Choices are
            ``["Embedding", "PadaDampG", "GeoDamp"]``.
        R : int (optional, default: ``1``)
            Adaptive sampling after ``R * n`` responses have been received.
        scorer : str (optional, default: ``"infogain"``)
            How queries should be scored. Scoring with ``scorer='infogain'``
            tries to link query score and "embedding improvement," and
            ``scorer='uncertainty'`` looks at the query that's closest to the
            decision boundary (or 50% probability).
        random_state : int, None, optional (default: ``None``)
            The random state to be used for initialization.
        kwargs : dict, optional
            Keyword arguments to pass to :class:`~salmon.triplets.samplers.adaptive.Embedding`.
        """
        super().__init__(ident=ident)

        logger.warning(f"Initializing Adaptive with n={n}, d={d}, R={R}")
        self.n = n
        self.d = d
        self.R = R

        self.n_search = kwargs.pop("n_search", 0)

        Opt = getattr(adaptive, optimizer)
        Module = getattr(adaptive, module)

        logger.info("Module = %s", Module)
        logger.info("opt = %s", Opt)

        self.opt = Opt(
            module=Module,
            module__n=n,
            module__d=d,
            module__random_state=random_state,
            optimizer=torch.optim.Adadelta,
            warm_start=True,
            max_epochs=200,
            **kwargs,
        )
        self.opt.initialize()

        probs = self.opt.net_.module_.probs
        if scorer == "infogain":
            search = InfoGainScorer(embedding=self.opt.embedding(), probs=probs)
        elif scorer == "uncertainty":
            search = UncertaintyScorer(embedding=self.opt.embedding(), probs=probs)
        else:
            raise ValueError(f"scorer={scorer} not in ['uncertainty', 'infogain']")

        self.search = search
        self.search.push([])
        self.meta = {
            "num_ans": 0,
            "model_updates": 0,
            "process_answers_calls": 0,
            "empty_pa_calls": 0,
        }
        self.params = {
            "n": n,
            "d": d,
            "R": R,
            "optimizer": optimizer,
            **kwargs,
        }

    def get_query(self, **kwargs) -> Tuple[Optional[Dict[str, int]], Optional[float]]:
        """Randomly select a query where there are few responses"""
        if self.meta["num_ans"] <= self.R * self.n:
            head, left, right = _random_query(self.n)
            return {"head": int(head), "left": int(left), "right": int(right)}, -9999
        return None, -9999

    def get_queries(
        self, num=None, stop=None, **kwargs
    ) -> Tuple[List[Query], List[float], dict]:
        """Get and score many queries."""
        ret_queries = []
        ret_scores = []
        n_searched = 0
        for pwr in range(12, 40 + 1):
            # I think there's a memory leak in search.score -- Dask workers
            # kept on dying on get_queries. min(pwr, 16) to fix that (and
            # verified too).
            #
            # pwr in range(12, 41) => about 1.7 million queries searched
            pwr = min(pwr, 16)
            queries, scores = self.search.score(num=2 ** pwr)
            n_searched += len(queries)
            ret_queries.append(queries)
            ret_scores.append(scores)

            # returned object is about (n_searched/1e6) * 16 MB in size
            # let's limit it to be 32MB in size
            if (n_searched >= 2e6) or (stop is not None and stop.is_set()):
                break
            if num or self.n_search:
                n_ret = int(num or self.n_search)
                if n_searched >= 3 * n_ret:
                    break

        queries = np.concatenate(ret_queries).astype(int)
        scores = np.concatenate(ret_scores)
        queries = self._sort_query_order(queries)

        ## Rest of this function takes about 450ms
        df = pd.DataFrame(queries)
        hashes = pd.util.hash_pandas_object(df, index=False)
        _, idx = np.unique(hashes.to_numpy(), return_index=True)
        queries = queries[idx]
        scores = scores[idx]
        if num or self.n_search:
            n_ret = int(num or self.n_search)
            queries = queries[:n_ret]
            scores = scores[:n_ret]
        return queries, scores, {}

    @staticmethod
    def _sort_query_order(queries: np.ndarray) -> np.ndarray:
        mins = np.minimum(queries[:, 1], queries[:, 2])
        maxs = np.maximum(queries[:, 1], queries[:, 2])
        queries[:, 1], queries[:, 2] = mins, maxs
        return queries

    def process_answers(self, answers: List[Answer]):
        """Process answers from the database.

        This function requires pulling from the database, and feeding those
        answers to the underlying optimization algorithm.
        """
        if not len(answers):
            self.meta["empty_pa_calls"] += 1
            if self.meta["empty_pa_calls"] >= 20:
                self.meta["empty_pa_calls"] = 0
                return self, True

        self.meta["num_ans"] += len(answers)
        self.meta["process_answers_calls"] += 1
        logger.debug("self.meta = %s", self.meta)
        logger.debug("self.R, self.n = %s, %s", self.R, self.n)

        # fmt: off
        alg_ans = [
            (a["head"], a["winner"],
             a["left"] if a["winner"] == a["right"] else a["right"])
            for a in answers
        ]
        # fmt: on
        self.search.push(alg_ans)
        self.search.embedding = self.opt.embedding()
        self.opt.push(alg_ans)
        if self.meta["num_ans"] < (self.R * self.n) / 10:
            return self, True

        # Make sure only valid answers are passed to partial_fit;
        # self.opt.answers_ has a bunch of extra space for new answers
        n_ans = self.opt.meta_["num_answers"]

        difficulty = np.log(self.params["n"]) * self.params["d"] * self.params["n"]
        if n_ans / difficulty <= 1:
            max_epochs = 200
        elif n_ans / difficulty <= 3:
            max_epochs = 120
        else:
            max_epochs = 50

        # max_epochs above for completely random initializations
        self.opt.set_params(max_epochs=max_epochs)

        valid_ans = self.opt.answers_[:n_ans]
        self.opt.fit(valid_ans)
        self.meta["model_updates"] += 1
        return self, True

    def get_model(self) -> Dict[str, Any]:
        """
        Get the embedding alongside other related information.
        """
        return {
            "embedding": self.search.embedding.tolist(),
            **self.meta,
            **self.params,
        }

    def predict(self, X, embedding=None):
        """
        Predict the answers of queries from the current embedding.

        Parameters
        ----------
        X : array-like
            Each row is ``[head, left, right]``. Each element in ``X`` or
            ``X[i, j]`` is an index of the current embedding.

        Returns
        -------
        y : array-like
            The winner of each query. An element of ``y`` is 0 if the left
            item is the predicted winner, and 1 if the right element is the
            predicted winner.

        """
        head_idx = X[:, 0].flatten()
        left_idx = X[:, 1].flatten()
        right_idx = X[:, 2].flatten()

        if embedding is None:
            embedding = self.opt.embedding()
        head = embedding[head_idx]
        left = embedding[left_idx]
        right = embedding[right_idx]

        ldiff = LA.norm(head - left, axis=1)
        rdiff = LA.norm(head - right, axis=1)

        # 1 if right closer; 0 if left closer
        # (which matches the labeling scheme)
        right_closer = rdiff < ldiff
        return right_closer.astype("uint8")

    def score(self, X, y, embedding=None):
        """
        Evaluate to see if current embedding agrees with the provided queries.

        Parameters
        ----------
        X : array-like, shape (n, 3)
            The columns should be aranged
        y : array-like, shape (n, )
            The answers to specific queries. The ``i``th value should be 0 if
            ``X[i, 1]`` won the query and 1 if ``X[i, 2]`` won the query.
        embedding : array-like, optional
            The embedding to use instead of the current embedding.
            The values in ``X`` will be treated as indices to this array.

        Returns
        -------
        acc : float
            The percentage of queries that agree with the current embedding.

        """
        y_hat = self.predict(X, embedding=embedding)
        y = np.asarray(y)
        y_hat = np.asarray(y_hat)
        return (y_hat == y).mean()


class TSTE(Adaptive):
    """The t-Distributed STE (t-STE) embedding algorithm [1]_.

    Notes
    -----
    This algorithm is proposed for the following reason:

    .. epigraph::

        In STE the value of the corresponding probability rapidly becomes
        infinitesimal when a triplet constraint is violated.  As a result,
        stronger violations of a constraint do not lead to significantly
        larger penalties, which reduces the tendency to correct triplet
        constraints that violate the consensus. This is illustrated by the
        STE gradient depicted in Figure 1:  the STE gradient is nearly zero
        when a constraint is strongly violated or satisfied. However, it
        appears that the gradient decays too rapidly, making it hard for
        STE to fix errors made early in the optimization later on.

        To address this problem, we propose to use a heavy-tailed kernel to
        measure local similarities between data points instead

        -- Section 4 of [1]_.


    References
    ----------
    .. [1] "Stochastic Triplet Embedding". 2012.
           http://www.cs.cornell.edu/~kilian/papers/stochastictriplet.pdf
           van der Maaten, Weinberger.
    """

    def __init__(self, alpha=1, **kwargs):
        """
        Parameters
        ----------
        alpha : float, default=1
            The parameter that controls how heavily the tails of the probability
            distribution are.
        kwargs : dict
            Keyword arguments to pass to :class:`~salmon.triplets.samplers.Adaptive`.
        """
        super().__init__(module="TSTE", module__alpha=alpha, **kwargs)


class ARR(Adaptive):
    """An adaptive "round robin" sampler.

    .. note::

       This class is usable in it's default configuration. Most of the
       :ref:`active sampling benchmarks <experiments>` have been run under
       the default configuration of this class. Please carefully consider
       any changes to the default parameters.

    Notes
    -----
    This algorithms asks about "high scoring queries" uniformly at random. For
    each head, the top ``n_top`` queries are selected. The query shown to the
    user is a query selected uniformly at random from this set.

    If ``n_top > 1``, then in practice, this sampling algorithm randomly asks
    about high scoring queries for each head. Becaues it's asynchronous, it
    randomly selects a head (instead of doing it a round-robin fashion).

    References
    ----------
    .. [1] Heim, Eric, et al. "Active perceptual similarity modeling with
           auxiliary information." `arXiv preprint arXiv:1511.02254
           <https://arxiv.org/abs/1511.02254>`_ (2015).

    """

    def __init__(
        self,
        R: int = 1,
        n_top: int = 1,
        module: str = "TSTE",
        priority: str = "random",
        **kwargs,
    ):

        """

        Parameters
        ----------
        R : int (optional, default ``1``)
            Adaptive sampling starts after ``R * n`` responses have been received.
        module : str, optional (default ``"TSTE"``).
            The noise model to use.
        n_top : int (optional, default ``1``)
            For each head, the number of top-scoring queries to ask about.
        priority : str, optional (default ``"random"``)
            Determines how queries should be ordered. Setting
            ``priority="random"`` will randomly shuffle queries; setting
            ``priority="original"`` will perserve the original scores. Setting
            ``priority="approx"`` will add some noise to the original score
            ranks.

            Regardless of the ``scores`` value, ``n_top`` queries per
            head will be perserved. It is likely most relevant when
            ``n_top==1``.

        kwargs : dict
            Keyword arguments to pass to :class:`~salmon.triplets.samplers.Adaptive`.

        """
        self.n_top = n_top
        if priority not in ["scores", "random", "approx"]:
            msg = f"priority={priority} not in ['random', 'original', 'approx']"
            raise ValueError(msg)
        self.priority = priority
        super().__init__(R=R, module=module, **kwargs)

    def get_queries(self, *args, **kwargs):
        queries, scores, meta = super().get_queries(*args, **kwargs)

        # (dataframe useful for manipulation below)
        df = pd.DataFrame(queries, columns=["h", "l", "r"])
        df["score"] = scores

        # Find the top scores per head
        top_scores_by_head = df.groupby(by="h")["score"].nlargest(n=self.n_top)
        top_idx = top_scores_by_head.index.droplevel(0)

        top_queries = df.loc[top_idx]
        top_scores = top_queries["score"].to_numpy()
        top_queries = top_queries.sample(frac=1, replace=False)

        posted = top_queries[["h", "l", "r"]].to_numpy().astype("int64")
        if self.priority == "random":
            r_scores = np.random.uniform(low=10, high=11, size=len(posted))
        elif self.priority == "scores":
            r_scores = top_queries["score"].to_numpy()
        elif self.priority == "approx":
            # ascending=True -> lowest to highest scores
            top_queries = top_queries.sort_values(by="score", ascending=True)
            r_scores = np.linspace(0, 1, num=len(top_queries))
            r_scores += 100 + np.random.uniform(0, 0.1, size=len(top_queries))
        else:
            msg = f"priority={self.priority} not in ['random', 'scores', 'approx']"
            raise ValueError(msg)

        meta.update({"n_queries_scored_(complete)": len(df)})
        return posted, r_scores, meta

    def process_answers(self, *args, **kwargs):
        new_self, updated = super().process_answers(*args, **kwargs)
        # Always return True to clear queries from the database (limits
        # randomness)
        return new_self, True


class SRR(ARR):
    """

    An adaptive round robin sampling strategy that performs a straighforward
    search; it performs a search of ``n_search`` queries with a randomly
    selected head.

    """

    def __init__(self, *args, n_search=400, **kwargs):
        """
        Parameters
        ----------
        n_search: int (optional, default ``400``)
            How many queries should be searched per user?
        kwargs : dict
            Keyword arguments to pass to :class:`~salmon.triplets.samplers.ARR`.
        """
        super().__init__(*args, **kwargs)
        self.n_search = n_search

    def get_queries(self, *args, **kwargs):
        return [], [], {}

    def get_query(self, **kwargs):
        q, score = super().get_query(**kwargs)
        if q is not None:
            return q, score

        head = int(np.random.choice(self.n))
        _choices = list(set(range(self.n)) - {head})
        choices = np.array(_choices)
        bottoms = [
            np.random.choice(choices, size=2, replace=False)
            for _ in range(self.n_search)
        ]

        _queries = [[head, l, r] for l, r in bottoms]
        queries, scores = self.search.score(queries=_queries)

        top_idx = np.argmax(scores)
        (h, l, r), score = queries[top_idx], float(scores[top_idx])
        return {"head": int(h), "left": int(l), "right": int(r)}, score


class STE(Adaptive):
    """The Stochastic Triplet Embedding [1]_.

    References
    ----------
    .. [1] "Stochastic Triplet Embedding". 2012.
           http://www.cs.cornell.edu/~kilian/papers/stochastictriplet.pdf
           van der Maaten, Weinberger.
    """

    def __init__(self, **kwargs):
        """
        Parameters
        ----------
        kwargs : dict
            Keyword arguments to pass to :class:`~salmon.triplets.samplers.Adaptive`.
        """
        super().__init__(module="STE", **kwargs)


class GNMDS(Adaptive):
    """The Generalized Non-metric Multidimensional Scaling embedding [1]_.

    References
    ----------
    .. [1] "Generalized Non-metric Multidimensional Scaling". 2007.
           Agarwal, Wills, Cayton, Lanckriet, Kriegman, and Belongie.
           http://proceedings.mlr.press/v2/agarwal07a/agarwal07a.pdf
    """

    def __init__(self, **kwargs):
        """
        Parameters
        ----------
        kwargs : dict
            Keyword arguments to pass to :class:`~salmon.triplets.samplers.Adaptive`.
        """
        super().__init__(module="GNMDS", **kwargs)


class CKL(Adaptive):
    """The crowd kernel embedding. Proposed in [1]_.

    References
    ----------
    .. [1] Tamuz, O., Liu, C., Belongie, S., Shamir, O., & Kalai, A. T. (2011).
           Adaptively learning the crowd kernel. https://arxiv.org/abs/1105.1033
    """

    def __init__(self, mu=1, **kwargs):
        """
        Parameters
        ----------
        mu : float
            The mu or :math:`\\mu` used in the CKL embedding. This is typically small; the default is :math:`10^{-4}`.
        kwargs : dict
            Keyword arguments to pass to :class:`~salmon.triplets.samplers.Adaptive`.
        """
        super().__init__(module__mu=mu, module="CKL", **kwargs)


class SOE(Adaptive):
    """The soft ordinal embedding detailed by Terada et al. [1]_

    This is evaluated as "SOE" by Vankadara et al., [2]_ in which they use the
    hinge loss on the distances (not squared distances).

    References
    ----------
    .. [1] Terada, Y. & Luxburg, U.. (2014). Local Ordinal Embedding.
           Proceedings of the 31st International Conference on Machine
           Learning, in PMLR 32(2):847-855.
           http://proceedings.mlr.press/v32/terada14.html

    .. [2] Vankadara, L. C., Haghiri, S., Lohaus, M., Wahab, F. U., &
           von Luxburg, U. (2019). Insights into Ordinal Embedding Algorithms:
           A Systematic Evaluation. https://arxiv.org/abs/1912.01666
    """

    def __init__(self, **kwargs):
        """
        Parameters
        ----------
        kwargs : dict
            Keyword arguments to pass to :class:`~salmon.triplets.samplers.Adaptive`.
        """
        super().__init__(module="SOE", **kwargs)



import itertools
import math
from time import time
from typing import Tuple, Union

import numpy as np
import numpy.linalg as LA
import torch

try:
    from salmon.triplets.samplers.adaptive.search import gram_utils
except:
    import gram_utils

Array = Union[np.ndarray, torch.Tensor]


def random_query(n):
    rng = np.random.RandomState()
    while True:
        a = rng.choice(n)
        b = rng.choice(n)
        c = rng.choice(n)
        if a != b and b != c and c != a:
            break
    return [a, b, c]


def embedding_matches(ans, D):
    """
    Parameters
    ----------
    ans = List[int], [head, winner, loser].
        Indices of the head/winner/loser
    D : np.ndarray
        Distance matrix

    Returns
    -------
    ans : int
        1 if agrees with D, 0 if doesn't
    """
    head, winner, loser = ans
    return D[head, winner] < D[head, loser]


def decide(D, head, winner, loser, distance=False, random=True):
    """
    Inputs
    ------
    D : np.ndarray
        Distance or embedding matrix.
        If embedding, n points in d dimensions. D.shape == (n, d)
        If distance, D[a, b] is distance between objects a and b.
    head, winner, loser: int
        Indices of head, winner and loser

    Returns
    -------
    prob : float
        The probability the triplet is satisfied
    """
    if distance:
        gram_utils.assert_distance(D)
        d_winner = D[head, winner]
        d_loser = D[head, loser]
    else:
        gram_utils.assert_embedding(D)
        d_winner = LA.norm(D[head] - D[winner]) ** 2
        d_loser = LA.norm(D[head] - D[loser]) ** 2
    if not random:
        q = [head, winner, loser] if d_winner < d_loser else [head, loser, winner]
        return q, {}

    # 0 < dist: agrees with embedding (d_loser > d_winner)
    dist = d_loser - d_winner  # >0: agrees with embedding. <0: does not agree.
    try:
        prob = 1 / (1 + np.exp(-dist))
    except FloatingPointError:
        prob = 0
    # d1 = d_winner, d2 = d_loser
    # = 1 / (1 + exp(-d2 + d1))
    # = 1 / (1 + exp(-d2 / -d1))
    # = exp(-d1) / (exp(-d1) + exp(d2))

    meta = {"prob": prob, "dist": dist}
    if np.random.rand() < prob:
        return [head, winner, loser], meta
    return [head, loser, winner], meta


def STE_probs(d1, d2, alpha=1):
    """
    Returns the probability that triplet wins
    """
    c = -(alpha + 1.0) / 2
    t1 = (1 + d1 / alpha) ** c
    t2 = (1 + d2 / alpha) ** c
    return t1 / (t1 + t2)


def exp_STE_probs(d2_winner, d2_loser):
    # dist>0: agrees with embedding. <0: does not agree.
    # d1 = d2_winner, d2 = d2_loser
    # 1 / (1 + exp(d1 - d2))
    # = 1 / (1 + exp(-d2 + d1))
    # = 1 / (1 + exp(-d2 / -d1))
    # = exp(-d1) / (exp(-d1) + exp(d2))
    # = prob of winning by STE
    return 1 / (1 + np.exp(d2_winner - d2_loser))


def entropy(x):
    if x.ndim == 1:
        i = x > 0
        y = x[i].copy()
        return (-1 * y * np.log(y)).sum()
    elif x.ndim == 2:
        i = np.argwhere(x > 0)
        y = x.copy()
        idx = (i[:, 0], i[:, 1])
        ret = -1 * y[idx] * np.log(y[idx])
        y[idx] = ret
        return np.sum(y, axis=1)
    else:
        raise ValueError("Invalid number of dimensions in input ``x``")


def posterior(D, S, alpha=1):
    gram_utils.assert_distance(D)
    n = D.shape[0]
    tau = np.zeros((n, n))
    for head, w, l in S:
        tau[head] += np.log(STE_probs(D[w], D[l], alpha=alpha))

    tau = np.exp(tau)
    s = tau.sum(axis=1)  # the sum of each row
    tau = (tau.T / s).T
    return tau


def getSTETripletProbability(i, j, k, alpha=1):
    """
    Return the probability of triplet [i,l,j] where a is closer to b than c.

    Namely:
    pabc = (1 + || c - a||^2/alpha)**(-(alpha+1)/2)/(2*alpha + || b - a ||^2+|| c - a ||^2)

    Inputs:
        (numpy.ndarray) a : numpy array
        (numpy.ndarray) b : numpy array
    (numpy.ndarray) c : numpy array
        (float) alpha : regularization parameter
    """
    ki = LA.norm(k - i)
    kj = LA.norm(k - j)
    c = -(alpha + 1.0) / 2
    return (1 + ki * ki / alpha) ** c / (
        (1 + ki * ki / alpha) ** c + (1 + kj * kj / alpha) ** c
    )


def posterior_orig(X, S, alpha=1):
    n = X.shape[0]
    tau = np.zeros((n, n))
    # Loop over each query
    for h, w, l in S:
        # Multiply by the amount the query contributes to each tau
        for i in range(n):
            tau[h, i] = tau[h, i] + math.log(
                getSTETripletProbability(X[w], X[l], X[i], alpha=alpha)
            )

    # Normalize -- make each row a PDF
    for a in range(n):
        tau[a] = np.exp(tau[a])
        s = sum(tau[a])
        tau[a] = tau[a] / s

    return tau


def posterior_embed(X, S, alpha=1):
    n = X.shape[0]
    tau = np.zeros((n, n))
    # Loop over each query
    for h, w, l in S:
        # Multiply by the amount the query contributes to each tau
        for i in range(n):
            tau[h, i] = tau[h, i] + math.log(
                getSTETripletProbability(X[w], X[l], X[i], alpha=alpha)
            )

    # Normalize -- make each row a PDF
    for i, a in enumerate(range(n)):
        ex = np.exp(tau[a])
        tau[a] = ex / ex.sum()

    return tau


def _score_v1(q: Tuple[int, int, int], tau: np.ndarray, D: np.ndarray) -> float:
    gram_utils.assert_distance(D)
    # head, o1, o2 = q
    head, w, l = q

    probs = STE_probs(D[w], D[l])  # probs.shape == (n, )
    eps = 1e-16
    # mask = (eps < np.abs(probs)) & (eps < np.abs(tau[head]))
    # mask = -1 < np.abs(probs)  # all probs

    p = (probs * tau[head]).sum()  # tau[head].shape == (n, )

    taub = tau[head] * probs
    taub /= taub.sum()

    tauc = tau[head] * (1 - probs)
    tauc /= tauc.sum()

    _entropy = -p * entropy(taub) - (1 - p) * entropy(tauc)
    return _entropy


def score(H: Array, W: Array, L: Array, tau: Array, D: Array, probs=STE_probs) -> Array:
    """
    Find the information gain for each query.

    Arguments
    ---------
    H, W, L : Array, Array, Array
        1D arrays describing the head and two choices respectively.
        If there are ``q`` questions, each array is of shape ``(q, )``.
    tau : Array
        Posterior, of shape ``(n, n)`` with ``n`` objects.
    D : Array
        Array of distances. Also of shape ``(n, n)``.

    Returns
    -------
    ig : Array of shape ``(q, )``
        The information gain of the queries (minus a constant).

    Notes
    -----

    The information gain of a query is given by the following expression:

    .. math::

       H(\tau) - pH(\tau_b) - (1 - p)H(\tau_a)

    where :math:`H` is the entropy. This

    References
    ----------
    [1] "Adaptively learning the crowd kernel" O. Tamuz, C. Liu,
         S. Belongie, O. Shamir, and A. Kalai. 2011.
         https://arxiv.org/abs/1105.1033

    """
    gram_utils.assert_distance(D)
    assert all(x.dtype.kind == "i" for x in (H, W, L))
    head, w, l = H, W, L
    q = len(head)

    probs = probs(D[l], D[w])  # (q, n)
    # probs = 1 - probs
    probs[np.isnan(probs)] = 0
    assert probs.min() >= 0
    eps = 0e-20
    # probs[probs < 1e-15] = 0

    p = (probs * tau[head]).sum(axis=1)  # (q, )

    _taub = tau[head] * probs  # (q, n)
    _taub[np.isnan(_taub)] = eps
    taub = _taub / (_taub.sum(axis=1).reshape(q, 1) + eps)

    _tauc = tau[head] * (1 - probs)  # (q, n)
    _tauc[np.isnan(_tauc)] = eps
    tauc = _tauc / (_tauc.sum(axis=1).reshape(q, 1) + eps)

    score = -p * entropy(taub) - (1 - p) * entropy(tauc)
    return score



import pytest

from salmon.triplets.manager import Config

html_keys = ["debrief", "instructions", "max_queries", "skip_button"]


@pytest.mark.parametrize("key", html_keys)
def test_old_html_keys_warns(key):
    c = Config()
    with pytest.raises(ValueError, match="Move.*into the `html` key"):
        c.parse({key: "foo"})


@pytest.mark.parametrize("key", html_keys)
def test_html_propogates(key):
    vals = {"max_queries": 50, "skip_button": True}
    v = vals.get(key, "foobar")

    user_config = {"html": {key: v}}
    c = Config().parse(user_config)

    assert c.html.dict()[key] == v


def test_defaults():
    user_config = {"targets": [0, 1, 2, 3]}
    c = Config().parse(user_config)

    assert c.dict() == {
        "targets": ["0", "1", "2", "3"],
        "html": {
            "instructions": "Please select the <i>comparison</i> item that is most similar to the <i>target</i> item.",
            "title": "Similarity judgements",
            "debrief": "<b>Thanks!</b> Please use the participant ID below.",
            "skip_button": False,
            "css": "",
            "max_queries": 50,
            "arrow_keys": True,
            "query_params": ["puid"],
            "js": "",
            "element_top": "",
            "element_middle": "",
            "element_bottom": "",
            "element_standalone": "",
        },
        "samplers": {"random": {"class": "Random"}},
        "sampling": {
            "common": {"d": 2},
            "probs": {"random": 100},
            "samplers_per_user": 0,
            "details": {},
        },
    }


def test_random_old_name_warns():
    user_config = {"samplers": {"RandomSampling": {}}}
    c = Config()
    with pytest.raises(ValueError, match="renamed to `Random`"):
        c.parse(user_config)


def test_wrong_probs():
    user_config = {
        "samplers": {"Random": {}, "ARR": {}},
        "sampling": {"probs": {"Random": 50, "ARR": 40}},
    }
    c = Config()
    with pytest.raises(ValueError, match="sampling.probs should add up to 100"):
        c.parse(user_config)


def test_wrong_keys():
    user_config = {
        "samplers": {"Random": {}, "ARR_v1": {}},
        "sampling": {"probs": {"Random": 50, "ARR_v2": 50}},
    }
    c = Config()
    with pytest.raises(ValueError, match="Keys in sampling.probs but not in samplers"):
        c.parse(user_config)


@pytest.mark.parametrize("v", [0, 1, 2, 3])
def test_bad_samplers_per_keys(v):
    c = Config()
    user_config = {
        "samplers": {"Random": {}, "ARR": {}},
        "sampling": {"samplers_per_user": v},
    }
    if v in [0, 1]:
        c.parse(user_config)
    else:
        with pytest.raises(
            NotImplementedError, match="Only samplers_per_user in {0, 1}"
        ):
            c.parse(user_config)


def test_d_default_in_common_params():
    user_config = {
        "samplers": {"ARR": {}},
        "sampling": {"common": {"foo": "bar"}},
    }
    c = Config().parse(user_config)
    assert c.sampling.common == {"d": 2, "foo": "bar"}


def test_d_default_updates():
    user_config = {
        "samplers": {"ARR": {}, "TSTE": {}},
        "sampling": {"common": {"d": 3}},
    }
    c = Config().parse(user_config)
    assert c.sampling.common == {"d": 3}

    user_config = {
        "samplers": {"ARR": {}, "TSTE": {}},
        "sampling": {"common": {"d": 3, "foo": "bar"}},
    }
    c = Config().parse(user_config)
    assert c.sampling.common == {"d": 3, "foo": "bar"}


def test_probs_default():
    user_config = {"samplers": {"ARR": {}, "TSTE": {}}}
    c = Config().parse(user_config)
    assert c.sampling.probs == {"ARR": 50, "TSTE": 50}


def test_sampling_details_no_sampler():
    samplers = {"ARR": {}, "TSTE": {}}
    details = {1: {"query": [1, 2, 3]}}
    config = {"samplers": samplers, "sampling": {"details": details}}
    with pytest.raises(
        ValueError,
        match="Specify the key `sampler` for each element of `sampling.details`",
    ):
        Config().parse(config)


def test_sampling_details_extra_sampler():
    samplers = {"ARR": {}, "TSTE": {}}
    details = {1: {"sampler": "fakeARR", "query": [1, 2, 3]}}
    config = {"samplers": samplers, "sampling": {"details": details}}
    with pytest.raises(
        ValueError,
        match="Each sampler specified in sampling.details must be created in samplers",
    ):
        Config().parse(config)


def test_sampling_details_bad_query_type():
    samplers = {"ARR": {}, "TSTE": {}}
    details = {1: {"sampler": "ARR", "query": {1: 2}}}
    config = {"samplers": samplers, "sampling": {"details": details}}
    with pytest.raises(
        ValueError, match="Not all `query` keys in sampling.details are lists"
    ):
        Config().parse(config)


def test_sampling_details_bad_lengths():
    samplers = {"ARR": {}, "TSTE": {}}
    details = {1: {"sampler": "ARR", "query": [1, 2, 3, 4]}}
    config = {"samplers": samplers, "sampling": {"details": details}}
    with pytest.raises(
        ValueError, match="Not all `query` keys in sampling.details have length 3"
    ):
        Config().parse(config)


def test_sampling_details_query_too_large():
    samplers = {"ARR": {}, "TSTE": {}}
    details = {1: {"sampler": "ARR", "query": [1, 2, 3, 4]}}
    config = {"samplers": samplers, "sampling": {"details": details}}
    with pytest.raises(
        ValueError, match="Not all `query` keys in sampling.details have length 3"
    ):
        Config().parse(config)


def test_sampling_details_query_too_large():
    samplers = {"ARR": {}, "TSTE": {}}
    details = {1: {"sampler": "ARR", "query": [10, 9, 8]}}
    config = {"targets": 10, "samplers": samplers, "sampling": {"details": details}}
    with pytest.raises(
        ValueError,
        match="Some target indices in the for sampling.details.query are too large.*\n\n.*0-based indexing",
    ):
        Config().parse(config)



import itertools
from copy import copy, deepcopy
from numbers import Number
from time import time
from typing import Dict, Union, List, Any

import numpy as np
import pandas as pd
import torch
import torch.optim as optim
from sklearn.base import BaseEstimator
from sklearn.exceptions import NotFittedError
from sklearn.model_selection import train_test_split

import salmon.triplets.samplers.adaptive as adaptive
from salmon.triplets.samplers.adaptive import OGD


def _get_params(opt_):
    return {
        k: v
        for k, v in opt_.get_params().items()
        if k[-1] != "_"
        and all(ignore not in k for ignore in ["callback", "iterator"])
        and k not in ["optimizer", "module", "criterion", "dataset"]
    }


def _print_fmt(v):
    if isinstance(v, (str, int)):
        return v
    if isinstance(v, (float, np.floating)):
        return f"{v:0.3f}"
    return v


def join(embedding: np.ndarray, targets: List[Any]) -> pd.DataFrame:
    em_final = pd.DataFrame(embedding)
    dims = {0: "x", 1: "y", 2: "z"}
    em_final.columns = [dims.get(c, f"dim{c + 1}") for c in em_final.columns]
    em_final["target"] = targets
    return em_final


class OfflineEmbedding(BaseEstimator):
    """
    Generate an embedding offline (after responses are downloaded from Salmon).

    Parameters
    ----------
    n : int
        Number of targets.
    d : int
        Embedding dimension
    max_epochs : int
        Number of epochs or passes through the dataset to run for.
    opt : Optional[Optimizer]
    verbose : int
        Interval at which to score.
    random_state : Optional[int]
        Random state for initialization.
    noise_model : str, optional (default: ``SOE``)
        Noise model to optimize over.
    kwargs : dict, optional, default: ``{}``
        Arguments for :class:`~salmon.triplets.offline.OGD`.
        Only used when ``opt is None``.

    """

    def __init__(
        self,
        n=None,
        d=None,
        max_epochs=400_000,
        opt=None,
        verbose=1000,
        ident="",
        noise_model="CKL",
        random_state=None,
        **kwargs,
    ):
        self.opt = opt
        self.n = n
        self.d = d
        self.max_epochs = max_epochs
        self.verbose = verbose
        self.ident = ident
        self.noise_model = noise_model
        self.random_state = random_state
        self.kwargs = kwargs

    @property
    def history_(self):
        """
        The history that's recorded during ``fit``. Available keys include
        ``score_test`` and ``loss_test``.
        """
        if self.initialized_:
            return self._history_
        raise NotFittedError(
            "No history has been recorded because ``initialize`` has not been called"
        )

    @property
    def meta_(self):
        """
        Meta-information about this estimator. Available keys include
        ``score_train`` and ``loss_train``.
        """
        return self._meta

    @property
    def embedding_(self):
        """
        The current embedding. If there are ``n`` objects being embedded into
        ``d`` dimensions, then ``embedding_.shape == (n, d)``.
        """
        return self.opt_.embedding_

    def initialize(self, X_train, embedding=None):
        """
        Initialize this optimizer.

        Parameters
        ----------
        X_train : np.ndarray
            Responses organized to be [head, winner, loser].

        embedding : nd.ndarray, optional
            If specified, initialize the embedding with the given values.

        """
        if self.opt is None:  # default
            assert self.n is not None and self.d is not None, "Specify n and d"
            noise_model = getattr(adaptive, self.noise_model)
            kwargs = dict(
                module=noise_model,
                module__n=self.n,
                module__d=self.d,
                module__random_state=self.random_state,
                optimizer=optim.Adadelta,
                max_epochs=self.max_epochs,
            )
            kwargs.update(self.kwargs)
            opt = OGD(**kwargs)
        else:
            opt = self.opt
        opt.initialize(embedding=embedding)
        opt.push(X_train)
        self._meta: Dict[str, Number] = {"pf_calls": 0}

        self.opt_ = opt
        self._history_ = []
        self.initialized_ = True
        return self

    def partial_fit(self, X_train):
        """
        Fit this optimizer for (approximately) one pass through the training data.

        Parameters
        ----------
        X_train : array-like
            The responses with shape ``(n_questions, 3)``.
            Each question is organized as ``[head, winner, loser]``.

        """
        if not (hasattr(self, "initialized_") and self.initialized_):
            self.initialize(X_train)

        self._partial_fit(X_train)
        return self

    def fit(self, X_train, X_test, embedding=None, get_stats=None, **stats_kwargs):
        """
        Fit the embedding with train and validation data.

        Parameters
        ----------
        X_train : array-like
            Data to fit the embedding too.

            The responses with shape ``(n_questions, 3)``.
            Each question is organized as ``[head, winner, loser]``.

        X_test : array-like
            Data to score the embedding on

            The responses with shape ``(n_questions, 3)``.
            Each question is organized as ``[head, winner, loser]``.

        embedding : np.ndarray, optional
            The embedding to initialize with.

            .. note::

               This is particularly useful when ``embedding`` is the
               online embedding from the CSV:

               .. code-block:: python

                  import pandas as pd
                  em = pd.read_csv("embedding.csv")  # from dashboard
                  df = pd.read_csv("responses.csv")  # from dashboard
                  X = df[["head", "winner", "loser"]]

                  from salmon.triplets.offline import OfflineEmbedding
                  est = OfflineEmbedding(...)
                  est.initialize(X, embedding=em)

        get_stats : Callable

        """
        self.initialize(X_train, embedding=embedding)
        self._meta["pf_calls"] = 0
        _start = time()
        if self.verbose:
            __score, __loss = self._score(X_test)
            print("initial score: ", __score)
            if get_stats:
                em = deepcopy(self.embedding_)
                datum2 = get_stats(embedding=em, X_test=X_test, **stats_kwargs)
                print("initial nn_diff_mean", datum2.get("nn_diff_mean", ""))
        for k in itertools.count():
            self._partial_fit(X_train)
            if self.verbose and k == 0:
                print(self.opt_.optimizer, self.opt_.get_params())

            if (self.verbose and k % self.verbose == 0) or abs(
                self.max_epochs - k
            ) <= 3:
                datum = deepcopy(self._meta)
                datum.update(self.opt_.meta_)
                for prefix, __X in [("train", X_train), ("test", X_test)]:
                    __score, __loss = self._score(__X)
                    datum[f"score_{prefix}"] = __score
                    datum[f"loss_{prefix}"] = __loss

                datum["_elapsed_time"] = int(time() - _start)

                if get_stats:
                    em = deepcopy(self.embedding_)
                    datum2 = get_stats(embedding=em, X_test=X_test, **stats_kwargs)
                    datum.update({f"stats__{k}": v for k, v in datum2.items()})

                self._history_.append(datum)
            if self.verbose and k % self.verbose == 0:
                # fmt: off
                keys = [
                    "ident", "score_test", "train_data", "max_epochs",
                    "_epochs", "_elapsed_time", "batch_size",
                ]
                # fmt: on
                show = {k: _print_fmt(datum.get(k, "")) for k in keys}
                print(show)
            if self.opt_.meta_["num_grad_comps"] >= self.max_epochs * len(X_train):
                break
        test_score, loss_test = self._score(X_test)
        self._history_[-1]["score_test"] = test_score
        self._history_[-1]["loss_test"] = loss_test
        return self

    def _score(self, X):
        module_ = self.opt_.module_
        with torch.no_grad():
            score = self.opt_.score(X)
            loss = module_.losses(*module_._get_dists(X))
        return score, float(loss.mean().item())

    def _partial_fit(self, X_train):
        _start = time()
        _n_ans = len(X_train)
        k = deepcopy(self._meta["pf_calls"])

        self.opt_.partial_fit(X_train)
        self._meta["pf_calls"] += 1
        self._meta.update(deepcopy(self.opt_.meta_))

        # train_score = self.opt_.score(X_train)
        # module_ = self.opt_.module_
        # loss_train = module_.losses(*module_._get_dists(X_train)).mean().item()

        prev_time = 0
        if len(self._history_):
            prev_time = self._history_[-1]["elapsed_time"]

        datum = {
            # "score_train": train_score,
            # "loss_train": loss_train,
            "k": k,
            "elapsed_time": time() - _start + prev_time,
            "train_data": len(X_train),
            "n": self.n,
            "d": self.d,
            "max_epochs": self.max_epochs,
            "verbose": self.verbose,
            "ident": self.ident,
        }
        datum["_epochs"] = self._meta["num_grad_comps"] / len(X_train)
        self._meta.update(datum)

        return self

    def score(self, X):
        """
        Score the responses against the current embedding. Record the loss and accuracy, and return the accuracy.

        Parameters
        ----------
        X : array-like

            The responses to score against the current embedding.

            The responses with shape ``(n_questions, 3)``.
            Each question is organized as ``[head, winner, loser]``.
        """
        acc, loss = self._score(X)
        self._meta["score__acc"] = acc
        self._meta["score__loss"] = loss
        return acc



import numpy as np
import numpy.linalg as LA
import pandas as pd
import pytest
from sklearn.utils import check_random_state

from salmon.triplets.samplers.adaptive import STE, TSTE, InfoGainScorer

from .. import _search as search
from .. import gram_utils
from . import utilsSTE


def test_probs():
    rng = check_random_state(0)
    n, d = 10, 2
    X = rng.randn(n, d)

    N = 100
    queries = [search.random_query(10) for _ in range(N)]
    p1 = [
        utilsSTE.getSTETripletProbability(X[o1], X[o2], X[h]) for h, o1, o2 in queries
    ]
    p1 = np.array(p1)

    d1 = np.array([LA.norm(X[h] - X[o1]) ** 2 for h, o1, o2 in queries])
    d2 = np.array([LA.norm(X[h] - X[o2]) ** 2 for h, o1, o2 in queries])
    p2 = search.STE_probs(d1, d2)
    assert np.allclose(p1, p2)

    p3 = search.exp_STE_probs(d1, d2)

    i = p2 > 0.5
    assert N / 3 <= i.sum() <= N * (1 - 1 / 3)
    assert (p3[i] > 0.5).all()


# @pytest.mark.xfail(reason="unknown bug (but negative scores provide adaptive gains)")
def test_score_refactor(seed=None):
    n, d = 40, 2
    rng = check_random_state(seed)
    X = rng.randn(n, d)
    G = X @ X.T
    D = gram_utils.distances(G)
    tau = rng.rand(n, n)
    for i in range(n):
        tau[i] /= tau[i].sum()

    queries = [search.random_query(n) for _ in range(100)]
    # old_score has been refactored to take in [h, w, l]
    old_scores = [_score_next([w, l, h], tau, X) for h, w, l in queries]

    Q = np.asarray(queries)
    H, W, L = Q[:, 0].flatten(), Q[:, 1].flatten(), Q[:, 2].flatten()
    new_scores = search.score(H, W, L, tau, D)

    old = np.array(old_scores)
    new = np.array(new_scores)
    assert np.allclose(old, new)


def test_probs_vector(seed=None):
    rng = check_random_state(seed)
    X = rng.randn(20, 2)
    n = X.shape[0]
    G = X @ X.T
    D = gram_utils.distances(G)

    for w, l in [(3, 5), (5, 3), (1, 4), (4, 1), (10, 15), (15, 10)]:
        p1 = [utilsSTE.getSTETripletProbability(X[w], X[l], X[i]) for i in range(n)]
        p2 = search.STE_probs(D[w], D[l])
        assert np.allclose(p1, p2)


def test_posterior(seed=None):
    """
    The two searches have different orderings for their queries.

    This depends on that ordering being correct. To resolve any differences,
    switch the ordering here. [[w, l, h] for h, w, l in S_D]
    """
    rng = check_random_state(seed)
    n, d = 40, 2
    X = rng.randn(n, d)
    G = X @ X.T
    D = gram_utils.distances(G)

    S_D = [search.random_query(n) for _ in range(200)]
    S_X = [[w, l, h] for h, w, l in S_D]

    new = search.posterior(D, S_D)
    old = utilsSTE.getSTETauDistribution(X, S_X)
    assert np.allclose(new, old)
    assert (new > 0).all()
    assert np.allclose(new.sum(axis=1), 1)


def test_posterior_v2(seed=None):
    """
    The two searches have different orderings for their queries.

    This depends on that ordering being correct. To resolve any differences,
    switch the ordering here. [[w, l, h] for h, w, l in S_D]
    """
    rng = check_random_state(seed)
    n, d = 40, 2
    X = rng.randn(n, d)
    G = X @ X.T
    D = gram_utils.distances(G)

    S_new = [search.random_query(n) for _ in range(200)]
    S_old = [[w, l, h] for h, w, l in S_new]

    new = search.posterior_embed(X, S_new)
    old = utilsSTE.getSTETauDistribution(X, S_old)
    assert np.allclose(new, old)
    assert (new > 0).all()
    assert np.allclose(new.sum(axis=1), 1)


def test_decide_distance(seed=None):
    rng = check_random_state(seed)
    n, d = 40, 2
    X = rng.randn(n, d)
    G = X @ X.T
    D = gram_utils.distances(G)

    queries = [search.random_query(n) for _ in range(50)]
    info1 = [search.decide(D, *q, distance=True, random=False) for q in queries]
    info2 = [search.decide(X, *q, distance=False, random=False) for q in queries]

    queries1 = [info[0] for info in info1]
    queries2 = [info[0] for info in info2]
    assert queries1 == queries2


def test_correct_prob(seed=None):
    rng = check_random_state(seed)
    n, d = 40, 2
    X = rng.randn(n, d)
    G = X @ X.T
    D = gram_utils.distances(G)

    queries = [search.random_query(n) for _ in range(10)]
    meta = [search.decide(X, *q)[1] for q in queries]
    queries = [search.decide(X, *q)[0] for q in queries]
    df = pd.DataFrame(meta)

    # cases where agrees with embedding X (distance more than 0)
    agrees = 0 < df.dist
    assert (0.5 < df["prob"][agrees]).all()

    # cases where disagrees with embedding (distance is less than 0)
    disagrees = df.dist < 0
    assert (df["prob"][disagrees] < 0.5).all()


def test_agrees_with_embedding(seed=None):
    rng = check_random_state(seed)
    n, d = 40, 2
    X = rng.randn(n, d)
    G = X @ X.T
    D = gram_utils.distances(G)

    # Make sure positive distance means query fits embedding
    queries = [search.random_query(n) for _ in range(10)]
    queries = [search.decide(X, *q, random=False)[0] for q in queries]
    meta = [search.decide(X, *q)[1] for q in queries]
    df = pd.DataFrame(meta)
    agrees = 0 < df.dist

    assert agrees.all()
    for head, winner, loser in queries:
        assert LA.norm(X[head] - X[winner]) < LA.norm(X[head] - X[loser])


def test_internal_search_refactor():
    """ Test to make sure Salmon search v1 and v2 produce the same results """
    n, d = 85, 2
    X = np.random.randn(n, d).astype("float32")
    G = gram_utils.gram_matrix(X)
    D = gram_utils.distances(G)
    tau = np.random.uniform(size=(n, n))

    Q = [
        (h, o1, o2)
        for h in range(n)
        for o1 in set(range(n)) - {h}
        for o2 in set(range(n)) - {h, o1}
    ]
    Q = Q[:30_000]
    s_old = [search._score_v1((h, o2, o1), tau, D) for h, o1, o2 in Q]

    Q = np.array(Q).astype(int)
    s_new = search.score(Q[:, 0], Q[:, 1], Q[:, 2], tau, D)
    assert np.allclose(s_new, np.array(s_old))


def _simple_triplet(n, rng, p_flip=0.1):
    h, a, b = rng.choice(n, size=3, replace=False)
    if abs(h - a) < abs(h - b):
        ret = [h, a, b]
    ret = [h, b, a]

    flip = rng.uniform(0, 1)
    if flip < p_flip:
        return [ret[0], ret[1], ret[0]]
    return ret


def test_salmon_integration():
    import warnings

    warnings.filterwarnings("error")
    n, d = 10, 2
    rng = np.random.RandomState(42)
    X = rng.randn(n, d).astype("float32")
    est = TSTE(n)
    search = InfoGainScorer(embedding=X, probs=est.probs)
    history = [_simple_triplet(n, rng) for _ in range(1000)]
    search.push(history)
    queries, scores = search.score()

    # This is the right order: queries are stored in next as (w, l, h). See myApp.py#L93.
    next_history = [[w, l, h] for h, w, l in history]
    tau = utilsSTE.getSTETauDistribution(X, next_history)

    rel_error = LA.norm(tau - search.posterior_) / LA.norm(tau)
    post = search.posterior_
    err = np.abs(search.posterior_ / tau)
    assert rel_error < 0.01

    # Making sure posterior approximately correct
    eps = 3e-3
    assert np.allclose(np.median(err), 1)
    assert 1 - eps <= np.percentile(err, 1) <= np.percentile(err, 99) <= 1 + eps

    # Make sure scores are all negative (because only negative entropy
    # part of information gain)
    next_scores = np.array([_score_next(q, tau, X) for q in next_history])
    _, salmon_scores = search.score(queries=history)
    assert (salmon_scores <= 0).all()

    # Make sure scores are close to each other
    rel_err = LA.norm(next_scores - salmon_scores) / LA.norm(next_scores)
    assert rel_err <= 1e-5

    diff = np.abs(next_scores - salmon_scores)
    assert 0 <= diff.min() <= diff.max() <= 2e-5
    assert diff.mean() <= 1e-5
    assert np.median(diff) <= 1e-5


def test_salmon_posterior_refactor(n=30, d=2):
    rng = np.random.RandomState(42)
    X = rng.randn(n, d).astype("float32")
    est = TSTE(n)
    search = InfoGainScorer(embedding=X, probs=est.probs)
    history = [_simple_triplet(n, rng) for _ in range(2000)]
    search.push(history)
    queries, scores = search.score()

    next_history = [[w, l, h] for h, w, l in history]
    tau = utilsSTE.getSTETauDistribution(X, next_history)

    rel_error = LA.norm(tau - search.posterior_) / LA.norm(tau)
    assert rel_error < 10e-6

    ratio = np.abs(search.posterior_ / tau)
    assert ratio.max() - ratio.min() < 0.1e-3
    assert pytest.approx(ratio.mean()) == 1


def _score_next(q: [int, int, int], tau, X):
    """
    copy/pasted from STE/myApp.py's getQuery
    (slightly modified)
    """
    n = X.shape[0]
    #  a, b, c = q
    # q = random_query(n) == [head, winner, loser]
    # utilsSTE.getRandomQuery == [winner, loser, head]
    # this is from utilsSTE.py#getRandomQuery and myAlg.py#get_query (which calls utilsSTE#getRandomQuery)
    # from myAlg.py#getQuery: b, c, a = q
    #                   (winner, loser, head)
    #  a, b, c = q
    b, c, a = q
    p = 0
    for i in range(n):
        p += utilsSTE.getSTETripletProbability(X[b], X[c], X[i]) * tau[a, i]

    taub = list(tau[a])
    for i in range(n):
        taub[i] = taub[i] * utilsSTE.getSTETripletProbability(X[b], X[c], X[i])
    taub = taub / sum(taub)

    tauc = list(tau[a])
    for i in range(n):
        tauc[i] = tauc[i] * utilsSTE.getSTETripletProbability(X[c], X[b], X[i])
    tauc = tauc / sum(tauc)

    entropy = -p * utilsSTE.getEntropy(taub) - (1 - p) * utilsSTE.getEntropy(tauc)
    return entropy



import asyncio
import logging
import logging.handlers
import os
from logging import LogRecord
from logging.handlers import QueueHandler
from pathlib import Path
# Python 3.7 and newer, fast reentrant implementation
# without task tracking (not needed for that when logging)
from queue import SimpleQueue as Queue
from typing import List

def get_logger(name):
    # Config from https://docs.python-guide.org/writing/logging/ and
    # https://docs.python-guide.org/writing/logging/
    logger = logging.getLogger(name)
    DEBUG = int(os.environ.get("SALMON_DEBUG", "0"))
    LOG_LEVEL = "DEBUG" if DEBUG else "INFO"
    LEVEL = getattr(logging, LOG_LEVEL)

    formatter = logging.Formatter("%(asctime)s %(name)s %(levelname)s: %(message)s")
    handlers = []

    ph = logging.StreamHandler()
    ph.setFormatter(formatter)
    ph.setLevel(LEVEL)
    handlers.append(ph)

    SRC = Path(__file__).absolute().parent
    assert SRC.is_dir()  # points to salmon/ source directory
    assert (SRC / "_out").exists(), str(SRC / "_out")
    out = SRC / "_out" / f"{name}.log"

    fh = logging.FileHandler(str(out))
    fh.setLevel(LEVEL)
    fh.setFormatter(formatter)
    handlers.append(fh)

    if False:
        # Works for uvicorn but not for gunicorn
        logger = background_logger(logger, *handlers)
    else:
        for handler in handlers:
            logger.addHandler(handler)
    logging.getLogger("fastapi").setLevel(LEVEL)
    return logger


def background_logger(logger, *handlers):
    """Move log handlers to a separate thread.

    Replace handlers on the root logger with a LocalQueueHandler,
    and start a logging.QueueListener holding the original
    handlers.

    Adapted https://www.zopatista.com/python/2019/05/11/asyncio-logging/
    """
    queue = Queue()
    async_handler = QueueHandler(queue)
    logger.addHandler(async_handler)

    listener = logging.handlers.QueueListener(
        queue, *handlers, respect_handler_level=True
    )
    listener.start()
    return logger


def flush_logger(logger):
    for handler in logger.handlers:
        try:
            handler.flush()
        except:
            pass
